import torch
from sklearn.metrics import mean_absolute_error, mean_squared_error, r2_score
import numpy as np
import matplotlib.pyplot as plt
import math
import argparse
import pickle
import os
import glob
from unicore.data import Dictionary

def get_result(filename):
    with open(filename, 'rb') as file:
        data = pickle.load(file)

    predict = []
    target = []
    src_tokens = []
    for item in data:
        predict.extend(item['predict'].reshape(-1).tolist())
        target.extend(item['target'].reshape(-1).tolist())
        # print(item.keys())
        src_token = item["src_token"][item["select_atom"]==1]
        src_token = src_token.detach().cpu().numpy().tolist()
        src_tokens.extend(src_token)
    return target, predict, src_tokens

def reg_metrics(target, predict):
    r2 = r2_score(target, predict)
    mae = mean_absolute_error(target, predict)
    mse = mean_squared_error(target, predict)
    rmse = math.sqrt(mse)
    return r2, mae, mse, rmse

def plot_metrics(target, predict, save_path=None, element="All"):
    r2, mae, mse, rmse = reg_metrics(target, predict)
    plt.figure(figsize=(8, 8))
    plt.scatter(target, predict, color='blue', alpha=0.5)
    xy_max = max(max(target), max(predict))
    xy_min = min(min(target), min(predict))
    plt.plot([xy_min, xy_max], [xy_min, xy_max], color='black', linestyle='--')
    plt.xlim(xy_min, xy_max)  
    plt.ylim(xy_min, xy_max)  
    plt.xlabel('Actual')
    plt.ylabel('Predicted')
    plt.title('Actual vs Predicted\nMAE: {:.4f}, RMSE: {:.4f}, R2: {:.4f}'.format(mae, rmse, r2))
    # plt.legend()
    if os.path.exists(save_path):
        fig_path = os.path.join(save_path, f'result_{element}.png')
        plt.savefig(fig_path)
    plt.show()


def main(args):
    dictionary = Dictionary.load(args.dict)
    if args.mode == 'cv':
        target = 0
        src_tokens = 0
        all_predict = []
        for folder in os.listdir(args.path):
            folder_path = os.path.join(args.path, folder)
            if os.path.isdir(folder_path):
                pkl_files = glob.glob(os.path.join(folder_path, "*.pkl"))
                filename = pkl_files[0]
                target, predict, src_tokens = get_result(filename)
                all_predict.append(predict)
                plot_metrics(target, predict, folder_path)
                r2, mae, mse, rmse = reg_metrics(target, predict)
                print(f'metric of {filename}\n: R2: {r2:.4f}, MAE: {mae:.4f}, MSE: {mse:.4f}, RMSE: {rmse:.4f}')

                elemenets = set(src_tokens)
                if len(elemenets) > 1:
                    for element in elemenets:
                        element_targets = np.array(target)[np.array(src_tokens)==element]
                        element_predicts = np.array(predict)[np.array(src_tokens)==element]
                        r2, mae, mse, rmse = reg_metrics(element_targets, element_predicts)
                        plot_metrics(target, predict, folder_path, dictionary[element])
                        print(f'metric of {dictionary[element]}\n: R2: {r2:.4f}, MAE: {mae:.4f}, MSE: {mse:.4f}, RMSE: {rmse:.4f}')

        if all_predict:
            mean_predict = np.mean(np.vstack(all_predict), axis=0)
            plot_metrics(target, mean_predict, args.path)
            r2, mae, mse, rmse = reg_metrics(target, mean_predict)
            print(f'metric of mean\n: R2: {r2:.4f}, MAE: {mae:.4f}, MSE: {mse:.4f}, RMSE: {rmse:.4f}')
            elemenets = set(src_tokens)
            if len(elemenets) > 1:
                for element in elemenets:
                    element_targets = np.array(target)[np.array(src_tokens)==element]
                    element_predicts = np.array(mean_predict)[np.array(src_tokens)==element]
                    r2, mae, mse, rmse = reg_metrics(element_targets, element_predicts)
                    plot_metrics(element_targets, element_predicts, args.path, dictionary[element])
                    print(f'metric of {dictionary[element]}\n: R2: {r2:.4f}, MAE: {mae:.4f}, MSE: {mse:.4f}, RMSE: {rmse:.4f}')

            mean_path = os.path.join(args.path, 'mean.pkl')
            mean_data = {'target': target, 'predict': mean_predict}
            with open(mean_path, 'wb') as file:
                pickle.dump(mean_data, file)

    else :
        pkl_files = glob.glob(os.path.join(args.path, "*.pkl"))
        filename = pkl_files[0]
        target, predict, src_tokens = get_result(filename)
        r2, mae, mse, rmse = reg_metrics(target, predict)
        print(f'metric of {filename}\n: R2: {r2:.4f}, MAE: {mae:.4f}, MSE: {mse:.4f}, RMSE: {rmse:.4f}')
        plot_metrics(target, predict, args.path)
        elemenets = set(src_tokens)
        if len(elemenets) > 1:
            for element in elemenets:
                element_targets = np.array(target)[np.array(src_tokens)==element]
                element_predicts = np.array(predict)[np.array(src_tokens)==element]
                r2, mae, mse, rmse = reg_metrics(element_targets, element_predicts)
                plot_metrics(target, predict, args.path, dictionary[element])
                print(f'metric of {dictionary[element]}\n: R2: {r2:.4f}, MAE: {mae:.4f}, MSE: {mse:.4f}, RMSE: {rmse:.4f}')


if __name__ == "__main__":

    parser = argparse.ArgumentParser(description="")
    parser.add_argument('--path', type=str)
    parser.add_argument('--mode', type=str)
    parser.add_argument('--dict', type=str)
    args = parser.parse_args()
    main(args)


# Copyright (c) DP Technology.
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

from unicore.data import Dictionary
import numpy as np

def parse_select_atom(dictionary, select_atom):
    
    if select_atom == 'All':
        return np.array([-1])
    else:
        return dictionary.vec_index(select_atom.split("&"))




# Copyright (c) DP Technology.
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

from unicore.data import BaseWrapperDataset
from functools import lru_cache
import torch

def copy_tensor(src, dst):
    assert dst.numel() == src.numel()
    dst.copy_(src)

def collate_tokens(
    values,
    pad_idx,
    left_pad=False,
    pad_to_length=None,
    pad_to_multiple=1,
):
    """Convert a list of 1d tensors into a padded 2d tensor."""
    size = max(v.size(0) for v in values)
    size = size if pad_to_length is None else max(size, pad_to_length)
    if pad_to_multiple != 1 and size % pad_to_multiple != 0:
        size = int(((size - 0.1) // pad_to_multiple + 1) * pad_to_multiple)
    res = values[0].new(len(values), size).fill_(pad_idx)

    for i, v in enumerate(values):
        copy_tensor(v, res[i][size - len(v) :] if left_pad else res[i][: len(v)])
    return res

def collate_tokens_2d(
    values,
    pad_idx,
    left_pad=False,
    pad_to_length=None,
    pad_to_multiple=1,
):
    """Convert a list of (batchsize, n, n) tensors into a padded (batchsize, size=max(n), size=max(n)) tensor."""
    size = max(v.size(0) for v in values)
    size = size if pad_to_length is None else max(size, pad_to_length)
    if pad_to_multiple != 1 and size % pad_to_multiple != 0:
        size = int(((size - 0.1) // pad_to_multiple + 1) * pad_to_multiple)
    res = values[0].new(len(values), size, size).fill_(pad_idx)

    for i, v in enumerate(values):
        copy_tensor(v, res[i][size - len(v):, size - len(v):] if left_pad else res[i][:len(v), :len(v)])
    return res

def collate_tokens_nm(
    values,
    pad_idx,
    left_pad=False,
    pad_to_length=None,
    pad_to_multiple=1,
):
    """Convert a list of (batchsize, n, m) tensors into a padded (batchsize, size=max(n), m) tensor."""
    size = max(v.size(0) for v in values)
    m = values[0].size(1)
    size = size if pad_to_length is None else max(size, pad_to_length)
    if pad_to_multiple != 1 and size % pad_to_multiple != 0:
        size = int(((size - 0.1) // pad_to_multiple + 1) * pad_to_multiple)
    res = values[0].new(len(values), size, m).fill_(pad_idx)
    
    for i, v in enumerate(values):
        copy_tensor(v, res[i][size - len(v) :,:] if left_pad else res[i][: len(v),:])
    return res

class PadDataset(BaseWrapperDataset):
    def __init__(self, dataset, pad_idx, left_pad):
        super().__init__(dataset)
        self.pad_idx = pad_idx
        self.left_pad = left_pad

    def collater(self, samples):
        return collate_tokens(samples, self.pad_idx, left_pad=self.left_pad, pad_to_multiple=8)


class LeftPadDataset(PadDataset):
    def __init__(self, dataset, pad_idx):
        super().__init__(dataset, pad_idx, left_pad=True)


class RightPadDataset(PadDataset):
    def __init__(self, dataset, pad_idx):
        super().__init__(dataset, pad_idx, left_pad=False)


class RightPadDataset2D(BaseWrapperDataset):
    def __init__(self, dataset, pad_idx,left_pad=False):
        super().__init__(dataset)
        self.pad_idx = pad_idx
        self.left_pad = left_pad
    def collater(self, samples):
        return collate_tokens_2d(samples, self.pad_idx, left_pad=self.left_pad, pad_to_multiple=8)
    
class RightPadDataset2D0(BaseWrapperDataset):
    def __init__(self, dataset, pad_idx,left_pad=False):
        super().__init__(dataset)
        self.pad_idx = pad_idx
        self.left_pad = left_pad
    def collater(self, samples):
        return collate_tokens_nm(samples, self.pad_idx, left_pad=self.left_pad, pad_to_multiple=8)

class PrependAndAppend2DDataset(BaseWrapperDataset):
    def __init__(self, dataset, token=None):
        super().__init__(dataset)
        self.token = token

    @lru_cache(maxsize=16)
    def __getitem__(self, idx):
        item = self.dataset[idx]
        if self.token is not None:
            h, w = item.size(-2), item.size(-1)
            new_item = torch.full((h+2, w+2), self.token).type_as(item)
            new_item[1:-1,1:-1] = item
            return new_item
        return item

class PrependAndAppend3DDataset(BaseWrapperDataset):
    def __init__(self, dataset, token=None):
        super().__init__(dataset)
        self.token = token

    @lru_cache(maxsize=16)
    def __getitem__(self, idx):
        item = self.dataset[idx]
        if self.token is not None:
            h, w, d = item.size(-3), item.size(-2), item.size(-1)
            new_item = torch.full((h+2, w+2, d), self.token).type_as(item)
            new_item[1:-1,1:-1,:] = item
            return new_item
        return item
    
def collate_tokens_3d(
    values,
    pad_idx,
    left_pad=False,
    pad_to_length=None,
    pad_to_multiple=1,
):
    """Convert a list of 3d tensors into a padded 3d tensor."""
    size = max(v.size(0) for v in values)
    d = values[0].size(-1)
    size = size if pad_to_length is None else max(size, pad_to_length)
    if pad_to_multiple != 1 and size % pad_to_multiple != 0:
        size = int(((size - 0.1) // pad_to_multiple + 1) * pad_to_multiple)
    res = values[0].new(len(values), size, size, d).fill_(pad_idx)

    def copy_tensor(src, dst):
        assert dst.numel() == src.numel()
        dst.copy_(src)

    for i, v in enumerate(values):
        copy_tensor(v, res[i][size - len(v):, size - len(v):, :] if left_pad else res[i][:len(v), :len(v), :])
    return res

class RightPadDataset3D(BaseWrapperDataset):
    def __init__(self, dataset, pad_idx,left_pad=False):
        super().__init__(dataset)
        self.pad_idx = pad_idx
        self.left_pad = left_pad
    def collater(self, samples):
        return collate_tokens_3d(samples, self.pad_idx, left_pad=self.left_pad, pad_to_multiple=8)


import re
from rdkit import Chem
from rdkit.Chem import AllChem
import warnings
from rdkit import RDLogger

import statistics
import numpy as np
from rdkit import Chem
from rdkit.Chem import AllChem
from multiprocessing import Pool
from write_lmdb import write_lmdb

import random
from collections import defaultdict
RDLogger.DisableLog('rdApp.*')
warnings.filterwarnings(action="ignore")



def get_element_symbols(mol):
    element_symbols = []  
    for atom in mol.GetAtoms():
        element_symbol = atom.GetSymbol() 
        element_symbols.append(element_symbol)
    return element_symbols

def process_single_molecule(args):
    db_id, value = args
    
    key = db_id + '_0'
    mol = nmr_data[key]['mol']
    smiles = nmr_data[key]['smiles']
    
    res = AllChem.EmbedMolecule(mol, randomSeed=seed)
    if res != 0:
        return None  

    try:
        AllChem.MMFFOptimizeMolecule(mol)
    except:
        return None   

    if mol.GetNumConformers() != 1:
        return None   

    atoms = get_element_symbols(mol)
    coordinates = np.array(mol.GetConformer().GetPositions())
    atoms_target = np.array([0.0] * len(atoms))
    atoms_target_mask = np.array([0] * len(atoms))

    index_shift_nmrdata = {}
    for dupmol_index in range(value+1):
        key = db_id + f'_{dupmol_index}'
        for nmrindex in range(len(nmr_data[key]['indexes_list'])):
            for indexes, shift in zip(nmr_data[key]['indexes_list'][nmrindex],nmr_data[key]['shifts_list'][nmrindex]):
                for index in indexes:
                    if index in index_shift_nmrdata:
                        index_shift_nmrdata[index].append(shift)
                    else:
                        index_shift_nmrdata[index] = [shift]

    for single_index, shifts in index_shift_nmrdata.items():
        atoms_target_mask[single_index-1] = 1
        if merge_method == "median":
            atoms_target[single_index-1] = statistics.median(shifts)
        elif merge_method == "mean":
            atoms_target[single_index-1] = statistics.mean(shifts)


    try:
        inchikey = Chem.MolToInchiKey(mol)
    except:
        inchikey = None

    return {
        'atoms': atoms,
        'coordinates': coordinates,
        'atoms_target': atoms_target,
        'atoms_target_mask': atoms_target_mask,
        'smiles': smiles,
        'db_id': db_id,
        'mol': mol,
        'inchikey': inchikey,
    }

file_path = '/vepfs/fs_users/xufanjie/data/NMR/dataset_nmrshiftdb2_2024/corrected_nmrshiftdb2.nmredata.sd'

supplier = Chem.ForwardSDMolSupplier(file_path, removeHs=False)    #

failed_mol_by_RDKit_count = 0
db_id_counter = {}
nmr_data = {}
for i, mol in enumerate(supplier):
    if mol is not None:    
        try:
            Chem.SanitizeMol(mol)
        except:
            continue
        if '.' in Chem.MolToSmiles(mol):
            continue

        smiles = mol.GetProp('NMREDATA_SMILES')

        db_id = mol.GetProp('NMREDATA_ID').split('=')[-1].replace('\\', '')
        if db_id not in db_id_counter:
            db_id_counter[db_id] = 0
        else:
            db_id_counter[db_id] += 1

        assignment_data = mol.GetPropsAsDict()['NMREDATA_ASSIGNMENT']
        assignment_pattern = r"s(\d+), ([-\d.]+), ((?:\d+, )*\d+)"
        assignment_data = re.findall(assignment_pattern, assignment_data)

        shifts_list = []
        indexes_list = []
        for prop_name in mol.GetPropNames():
            if prop_name.startswith('NMREDATA_1D_'):

                nmr_1d_data_raw = mol.GetPropsAsDict()[prop_name]
                
                nmr_1d_spectrum_pattern = r"([-\d.]+), L=s(\d+)"
                nmr_1d_data = re.findall(nmr_1d_spectrum_pattern, nmr_1d_data_raw)
                

                single_shifts_list = []
                single_indexes_list = []
                isDirty = False
                for shift, peak_index in nmr_1d_data:
                    isDirty = False
                    peak_index = int(peak_index)
                    
                    assignment_index = next((j for j, k in enumerate(assignment_data) if int(k[0]) == peak_index), None)
                    if assignment_index is None:
                        continue
                    atom_numbers = [int(x) for x in assignment_data[assignment_index][2].split(', ')]

                    if float(shift) != float(assignment_data[assignment_index][1]):
                        print(f"{db_id}_{db_id_counter[db_id]}: In molecule {prop_name}, the chemical shifts differ when the assignment index matches the peak number in 1D NMR: 1D NMR shift: {float(shift)}, assignment shift: {float(assignment_data[assignment_index][1])}")
                        continue
                    
                    nmr_1d_type_pattern = re.compile(r'NMREDATA_1D_\d+([A-Z]+)')
                    nmr_1d_type = re.findall(nmr_1d_type_pattern, prop_name)
                    for atom_index in atom_numbers:
                        atom = mol.GetAtomWithIdx(atom_index-1)
                        element_type = atom.GetSymbol()
                        if element_type != nmr_1d_type[0]:
                            isDirty = True
                            break
                    
                    if isDirty:
                        break

                    if nmr_1d_type[0] == 'H':
                        if float(shift) > 15 or  float(shift) < 0:
                            isDirty = True
                            break
                    elif nmr_1d_type[0] == 'C':
                        if float(shift) > 250 or  float(shift) < -50:
                            isDirty = True
                            break
                    elif nmr_1d_type[0] == 'F':
                        if float(shift) > 100 or  float(shift) < -250:
                            isDirty = True
                            break

                    if len(atom_numbers) != len(set(atom_numbers)):
                        print(f"{db_id}_{db_id_counter[db_id]}: In molecule {prop_name}, the same chemical shift corresponds to multiple identical indices, possibly due to incorrect numbering of equivalent atoms.")

                    single_shifts_list.append((float(shift)))
                    single_indexes_list.append(atom_numbers)


                if isDirty:
                    break
                else:
                    if ((len(single_shifts_list) == 0) or (len(single_indexes_list) == 0)):
                        continue
                    else:
                        shifts_list.append(single_shifts_list)
                        indexes_list.append(single_indexes_list)

        if ((len(shifts_list) == 0) or (len(indexes_list) == 0)):
            db_id_counter[db_id] -= 1
            if (db_id_counter[db_id] == -1):
                del db_id_counter[db_id]
            continue

        key = f'{db_id}_{db_id_counter[db_id]}' 
        nmr_data[key] = {
            'smiles': smiles,
            'mol': mol,
            'db_id': db_id,
            'shifts_list': shifts_list,
            'indexes_list': indexes_list
        }
    else:
        failed_mol_by_RDKit_count += 1



merge_method = "median"  # "median" or "mean"

non_merged_duplicate_index = -1

seed = 42

tasks = [(db_id, value) for db_id, value in db_id_counter.items()]

with Pool() as pool:
    results = pool.map(process_single_molecule, tasks)

nmrdata_list = [result for result in results if result is not None]


random.seed(42)

inchikey_groups = defaultdict(list)
for idx, item in enumerate(nmrdata_list):
    inchikey = item['inchikey']
    inchikey_groups[inchikey].append(idx)

duplicate_indices_list = [idx for indices in inchikey_groups.values() if len(indices) > 1 for idx in indices]

non_duplicate_indices = [idx for indices in inchikey_groups.values() if len(indices) == 1 for idx in indices]

random.shuffle(non_duplicate_indices)  
split_index = int(len(nmrdata_list) * 0.8) - len(duplicate_indices_list)   

train_index = non_duplicate_indices[:split_index] + duplicate_indices_list
test_index = non_duplicate_indices[split_index:]

train_data = [nmrdata_list[i] for i in train_index]
test_data = [nmrdata_list[i] for i in test_index]

valid_elements = {'C', 'H', 'O', 'N', 'F', 'B'}

train_data = []
for index in train_index:
    skip_molecule = False
    for atom_index, atom in enumerate(nmrdata_list[index]['atoms']):
        if atom not in valid_elements:
            nmrdata_list[index]['atoms_target_mask'][atom_index] = 0
        
        # Additional conditions for N and O
        if atom == 'N' and nmrdata_list[index]['atoms_target'][atom_index] < -100:
            skip_molecule = True
            break
        if atom == 'O' and nmrdata_list[index]['atoms_target'][atom_index] > 490:
            skip_molecule = True
            break

    if all(element == 0 for element in nmrdata_list[index]['atoms_target_mask']):
        print(f"{index} molecule has uncommon active elements")
    elif skip_molecule:
        print(f"{index} molecule has uncommon NMR")
    else:
        train_data.append(nmrdata_list[index])

test_data = []
for index in test_index:
    skip_molecule = False
    for atom_index, atom in enumerate(nmrdata_list[index]['atoms']):
        if atom not in valid_elements:
            nmrdata_list[index]['atoms_target_mask'][atom_index] = 0
        
        # Additional conditions for N and O
        if atom == 'N' and nmrdata_list[index]['atoms_target'][atom_index] < -100:
            skip_molecule = True
            break
        if atom == 'O' and nmrdata_list[index]['atoms_target'][atom_index] > 490:
            skip_molecule = True
            break

    if all(element == 0 for element in nmrdata_list[index]['atoms_target_mask']):
        print(f"{index} molecule has uncommon active elements")
    elif skip_molecule:
        print(f"{index} molecule has uncommon NMR")
    else:
        test_data.append(nmrdata_list[index])

ele = 'All'
mode = 'train'
write_lmdb(f'/vepfs/fs_users/xufanjie/NMRNet/data/nmrshiftdb2_2024/{ele}/{mode}.lmdb', train_data)
mode = 'valid'
write_lmdb(f'/vepfs/fs_users/xufanjie/NMRNet/data/nmrshiftdb2_2024/{ele}/{mode}.lmdb', test_data)


# Copyright (c) DP Technology.
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

import numpy as np
import torch
import joblib
import os
from sklearn.preprocessing import (
    StandardScaler,
    MinMaxScaler,
    MaxAbsScaler,
    RobustScaler,
    Normalizer,
    QuantileTransformer,
    PowerTransformer,
    FunctionTransformer,
)

SCALER_MODE = {
    'minmax': MinMaxScaler(),
    'standard': StandardScaler(),
    'robust': RobustScaler(),
    'maxabs': MaxAbsScaler(),
    'quantile': QuantileTransformer(),
    'power_box': PowerTransformer(method='box-cox'),
    'power_yeo': PowerTransformer(method='yeo-johnson'),
    'normalizer': Normalizer(),
    'log1p': FunctionTransformer(np.log1p),
}

class TargetScaler(object):
    def __init__(self, load_dir=None, scaler_file = 'target_scaler.ss'):
        if load_dir and os.path.exists(os.path.join(load_dir, scaler_file)):
            self.scaler = joblib.load(os.path.join(load_dir, scaler_file))
        else:
            self.scaler = None

    def fit(self, target, num_classes, dump_dir):
        if os.path.exists(os.path.join(dump_dir, 'target_scaler.ss')):
            self.scaler = joblib.load(os.path.join(dump_dir, 'target_scaler.ss'))
        else:
            self.scaler = SCALER_MODE['standard']
            target_selected = target.reshape(-1, num_classes)
            self.scaler.fit(target_selected)

            joblib.dump(self.scaler, os.path.join(dump_dir, 'target_scaler.ss'))

    def transform(self, target):
        return self.scaler.transform(target)


    def inverse_transform(self, target):
        return self.scaler.inverse_transform(target)

    # def fit(self, target, num_classes, dump_dir, mask=None):
    #     if os.path.exists(os.path.join(dump_dir, 'target_scaler.ss')):
    #         return
    #     else:
    #         self.scaler = SCALER_MODE['standard']
    #         if mask == None:
    #             mask = np.ones_like(target)
    #         target_selected = target[mask==1].reshape(-1, num_classes)
    #         self.scaler.fit(target_selected)

    # def transform(self, target, num_classes, mask=None):
    #     if mask == None:
    #         mask = torch.ones_like(target)
    #     return self.scaler.transform(target[mask==1].reshape(-1, num_classes)).reshape(target.shape)


    # def inverse_transform(self, target, num_classes, mask=None):
    #     if mask == None:
    #         mask = torch.ones_like(target)
    #     return self.scaler.inverse_transform(target[mask==1].reshape(-1, num_classes)).reshape(target.shape)







# Copyright (c) DP Technology.
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

from functools import lru_cache

import numpy as np
import torch
from unicore.data import Dictionary, data_utils
from unicore.data import BaseWrapperDataset


class MaskPointsChunkDataset(BaseWrapperDataset):

    def __init__(
        self,
        dataset: torch.utils.data.Dataset,
        coord_dataset: torch.utils.data.Dataset,
        vocab: Dictionary,
        pad_idx: int,
        mask_idx: int,
        noise_type: str,
        noise: float = 1.0,
        seed: int = 1,
        mask_prob: float = 0.15,
        leave_unmasked_prob: float = 0.1,
        random_token_prob: float = 0.1,
        chunk_mask_prob: float = 0.5,
        chunk_ratio: float = 0.5,
    ):
        assert 0.0 < mask_prob < 1.0
        assert 0.0 <= random_token_prob <= 1.0
        assert 0.0 <= leave_unmasked_prob <= 1.0
        assert random_token_prob + leave_unmasked_prob <= 1.0
        assert 0.0 <= chunk_mask_prob <= 1.0

        self.dataset = dataset
        self.coord_dataset = coord_dataset
        self.vocab = vocab
        self.pad_idx = pad_idx
        self.mask_idx = mask_idx
        self.noise_type = noise_type
        self.noise = noise
        self.seed = seed
        self.mask_prob = mask_prob
        self.leave_unmasked_prob = leave_unmasked_prob
        self.random_token_prob = random_token_prob
        self.chunk_mask_prob = chunk_mask_prob
        self.chunk_ratio = chunk_ratio

        if random_token_prob > 0.0:
            weights = np.ones(len(self.vocab))
            weights[vocab.special_index()] = 0
            self.weights = weights / weights.sum()

        self.epoch = None
        if self.noise_type == 'trunc_normal':
            self.noise_f = lambda num_mask: np.clip(np.random.randn(num_mask, 3) * self.noise, a_min=-self.noise*2.0, a_max=self.noise*2.0)
        elif self.noise_type == 'normal':
            self.noise_f = lambda num_mask: np.random.randn(num_mask, 3) * self.noise
        elif self.noise_type == 'uniform':
            self.noise_f = lambda num_mask: np.random.uniform(low=-self.noise, high=self.noise, size=(num_mask, 3))
        else:
            self.noise_f = lambda num_mask: 0.0

    def set_epoch(self, epoch, **unused):
        super().set_epoch(epoch)
        self.coord_dataset.set_epoch(epoch)
        self.dataset.set_epoch(epoch)
        self.epoch = epoch

    def __getitem__(self, index: int):
        return self.__getitem_cached__(self.epoch, index)

    @lru_cache(maxsize=16)
    def __getitem_cached__(self, epoch: int, index: int):
        ret = {}
        with data_utils.numpy_seed(self.seed, epoch, index):
            item = self.dataset[index]
            coord = self.coord_dataset[index]
            sz = len(item)
            # don't allow empty sequence
            assert sz > 0
            if np.random.rand() > self.rand_mask_prob:
                ##### random mask tokens and noise coordinates #####
                # decide elements to mask
                num_mask = int(
                    # add a random number for probabilistic rounding
                    self.mask_prob * sz + np.random.rand()
                )
                mask_idc = np.random.choice(sz, num_mask, replace=False)
                mask = np.full(sz, False)
                mask[mask_idc] = True
                ret['targets'] = np.full(len(mask), self.pad_idx)
                ret['targets'][mask] = item[mask]
                ret['targets'] = torch.from_numpy(ret['targets']).long()
                # decide unmasking and random replacement
                rand_or_unmask_prob = self.random_token_prob + self.leave_unmasked_prob
                if rand_or_unmask_prob > 0.0:
                    rand_or_unmask = mask & (np.random.rand(sz) < rand_or_unmask_prob)
                    if self.random_token_prob == 0.0:
                        unmask = rand_or_unmask
                        rand_mask = None
                    elif self.leave_unmasked_prob == 0.0:
                        unmask = None
                        rand_mask = rand_or_unmask
                    else:
                        unmask_prob = self.leave_unmasked_prob / rand_or_unmask_prob
                        decision = np.random.rand(sz) < unmask_prob
                        unmask = rand_or_unmask & decision
                        rand_mask = rand_or_unmask & (~decision)
                else:
                    unmask = rand_mask = None

                if unmask is not None:
                    mask = mask ^ unmask

                new_item = np.copy(item)
                new_item[mask] = self.mask_idx

                num_mask = mask.astype(np.int32).sum()
                new_coord = np.copy(coord)
                new_coord[mask, :] += self.noise_f(num_mask)

                if rand_mask is not None:
                    num_rand = rand_mask.sum()
                    if num_rand > 0:
                        new_item[rand_mask] = np.random.choice(
                            len(self.vocab),
                            num_rand,
                            p=self.weights,
                        )
                ret['atoms'] = torch.from_numpy(new_item).long()
                ret['coordinates'] = torch.from_numpy(new_coord).float()
            else:
                #### chunk mask based on coordinates ####
                minx, miny, minz = coord.min(axis=0)
                maxx, maxy, maxz = coord.max(axis=0)
                


            return ret


# Copyright (c) DP Technology.
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

import logging
import os

import numpy as np
import torch
from unicore.data import (
    Dictionary,
    NestedDictionaryDataset,
    AppendTokenDataset,
    PrependTokenDataset,
    RightPadDataset,
    EpochShuffleDataset,
    TokenizeDataset,
    RightPadDataset2D,
    LMDBDataset
)
from ..data import (
    KeyDataset,
    ToTorchDataset,
    MaskPointsDataset,
    DistanceDataset,
    GlobalDistanceDataset,
    EdgeTypeDataset,
    RightPadDataset3D,
    PrependAndAppend2DDataset,
    PrependAndAppend3DDataset,
    RightPadDataset2D0,
    LatticeNormalizeDataset,
    LatticeMatrixNormalizeDataset,
    RemoveHydrogenDataset,
    CroppingDataset,
    NormalizeDataset,
    SelectTokenDataset,
    FilterDataset,
)
from unicore.tasks import UnicoreTask, register_task
from ...uninmr.utils import parse_select_atom


logger = logging.getLogger(__name__)


@register_task("unimat")
class UniMatTask(UnicoreTask):
    """Task for training transformer auto-encoder models."""

    @staticmethod
    def add_args(parser):
        """Add task-specific arguments to the parser."""
        parser.add_argument(
            "data",
            help="colon separated path to data directories list, \
                            will be iterated upon during epochs in round-robin manner",
        )
        parser.add_argument(
            "--mask-prob",
            default=0.15,
            type=float,
            help="probability of replacing a token with mask",
        )
        parser.add_argument(
            "--leave-unmasked-prob",
            default=0.05,
            type=float,
            help="probability that a masked token is unmasked",
        )
        parser.add_argument(
            "--random-token-prob",
            default=0.05,
            type=float,
            help="probability of replacing a token with a random token",
        )
        parser.add_argument(
            "--noise-type",
            default="uniform",
            choices=["trunc_normal", "uniform", "normal", "none"],
            help="noise type in coordinate noise",
        )
        parser.add_argument(
            "--noise",
            default=1.0,
            type=float,
            help="coordinate noise for masked atoms",
        )
        parser.add_argument(
            "--max-atoms",
            type=int,
            default=512,
            help="selected maximum number of atoms in a molecule",
        )
        parser.add_argument(
            "--dict-name",
            default="dict.txt",
            help="dictionary file",
        )
        parser.add_argument(
            "--dist-threshold",
            type=float,
            default=8.0,
            help="distance threshold for distance loss",
        )
        parser.add_argument(
            "--minkowski-p",
            type=float,
            default=2.0,
            help="minkowski p for distance loss",
        )
        parser.add_argument(
            "--random-choice",
            type=float,
            default=-1.0,
            help="random choice atom to predict lattice",
        )
        parser.add_argument(
            "--remove-hydrogen",
            action="store_true",
            help="remove hydrogen atoms",
        )
        parser.add_argument(
            "--global-distance",
            action="store_true",
            help="use global distance",
        )
        parser.add_argument(
            "--selected-atom",
            default="All",
            help="select atom: All or H or H&C&F...",
        )

    def __init__(self, args, dictionary):
        super().__init__(args)
        self.dictionary = dictionary
        self.seed = args.seed
        # add mask token
        self.mask_idx = dictionary.add_symbol("[MASK]", is_special=True)
        self.selected_token = parse_select_atom(self.dictionary, args.selected_atom)

    @classmethod
    def setup_task(cls, args, **kwargs):
        dictionary = Dictionary.load(os.path.join(args.data, args.dict_name))
        logger.info("dictionary: {} types".format(len(dictionary)))
        return cls(args, dictionary)

    def load_dataset(self, split, combine=False, **kwargs):
        """Load a given dataset split.
        Args:
            split (str): name of the split (e.g., train, valid, test)
        """
        split_path = os.path.join(self.args.data, split + ".lmdb")
        dataset = LMDBDataset(split_path)
        if self.args.remove_hydrogen:
            dataset = RemoveHydrogenDataset(dataset, 'atoms', 'coordinates')
        dataset = CroppingDataset(dataset, self.args.seed, 'atoms', 'coordinates', self.args.max_atoms)
        dataset = NormalizeDataset(dataset, 'coordinates')

        token_dataset = KeyDataset(dataset, "atoms")
        token_dataset = TokenizeDataset(token_dataset, self.dictionary, max_seq_len=self.args.max_seq_len)
        # atoms_target_mask_dataset = KeyDataset(dataset, "atoms_target_mask")
        select_atom_dataset = SelectTokenDataset(token_dataset=token_dataset, selected_token=self.selected_token, random_choice=self.args.random_choice)

        coord_dataset = KeyDataset(dataset, "coordinates")

        expand_dataset = MaskPointsDataset(
            token_dataset,
            coord_dataset,
            self.dictionary,
            pad_idx=self.dictionary.pad(),
            mask_idx=self.mask_idx,
            noise_type=self.args.noise_type,
            noise=self.args.noise,
            seed=self.args.seed,
            mask_prob=self.args.mask_prob,
            leave_unmasked_prob=self.args.leave_unmasked_prob,
            random_token_prob=self.args.random_token_prob,
        )

        def PrependAndAppend(dataset, pre_token, app_token):
            dataset = PrependTokenDataset(dataset, pre_token)
            return AppendTokenDataset(dataset, app_token)

        encoder_token_dataset = KeyDataset(expand_dataset, "atoms")
        encoder_target_dataset = KeyDataset(expand_dataset, "targets")
        encoder_coord_dataset = KeyDataset(expand_dataset, "coordinates")

        token_dataset = PrependAndAppend(token_dataset, self.dictionary.bos(), self.dictionary.pad())
        select_atom_dataset = PrependAndAppend(select_atom_dataset, self.dictionary.bos(), self.dictionary.pad())

        src_dataset = PrependAndAppend(encoder_token_dataset, self.dictionary.bos(), self.dictionary.eos())
        tgt_dataset = PrependAndAppend(encoder_target_dataset, self.dictionary.pad(), self.dictionary.pad())
        coord_dataset = ToTorchDataset(coord_dataset, 'float32')

        if self.args.global_distance:
            lattice_dataset = LatticeNormalizeDataset(dataset, 'abc', 'angles')
            lattice_dataset = ToTorchDataset(lattice_dataset, 'float32')
            lattice_matrix_dataset = LatticeMatrixNormalizeDataset(dataset, 'lattice_matrix')
            logger.info("use global distance: {}".format(self.args.global_distance))
            encoder_distance_dataset = GlobalDistanceDataset(encoder_coord_dataset, lattice_matrix_dataset, p=self.args.minkowski_p)
            encoder_distance_dataset = PrependAndAppend3DDataset(encoder_distance_dataset, 0.0)
            encoder_distance_dataset = RightPadDataset3D(encoder_distance_dataset, pad_idx=0)

            distance_dataset = GlobalDistanceDataset(coord_dataset, lattice_matrix_dataset, p=self.args.minkowski_p)
            distance_dataset = PrependAndAppend3DDataset(distance_dataset, 0.0)
            distance_dataset = RightPadDataset3D(distance_dataset, pad_idx=0)
        else:
            lattice_dataset = None
            encoder_distance_dataset = DistanceDataset(encoder_coord_dataset, p=self.args.minkowski_p)
            encoder_distance_dataset = PrependAndAppend2DDataset(encoder_distance_dataset, 0.0)
            encoder_distance_dataset = RightPadDataset2D(encoder_distance_dataset, pad_idx=0)

            distance_dataset = DistanceDataset(coord_dataset, p=self.args.minkowski_p)
            distance_dataset = PrependAndAppend2DDataset(distance_dataset, 0.0)
            distance_dataset = RightPadDataset2D(distance_dataset, pad_idx=0)

        encoder_coord_dataset = PrependAndAppend(encoder_coord_dataset, 0.0, 0.0)
        coord_dataset = PrependAndAppend(coord_dataset, 0.0, 0.0)

        edge_type = EdgeTypeDataset(src_dataset, len(self.dictionary))

        net_input = {
                "select_atom": RightPadDataset(
                    select_atom_dataset,
                    pad_idx=self.dictionary.pad(),
                ),
                "src_tokens": RightPadDataset(
                    src_dataset,
                    pad_idx=self.dictionary.pad(),
                ),
                "src_coord": RightPadDataset2D0(
                    encoder_coord_dataset,
                    pad_idx=0,
                ),
                "src_distance": encoder_distance_dataset,
                "src_edge_type": RightPadDataset2D(
                    edge_type,
                    pad_idx=0,
                ),
        }
        target = {
            "tokens_target": RightPadDataset(tgt_dataset, pad_idx=self.dictionary.pad()),
            "distance_target": distance_dataset,
            "coord_target": RightPadDataset2D0(coord_dataset, pad_idx=0),
            "lattice_target": lattice_dataset,
            }
        dataset = {"net_input": net_input, "target": target}
        dataset = NestedDictionaryDataset(
            dataset
        )
        if split in ["train", "train.small"]:
            dataset = EpochShuffleDataset(dataset, len(dataset), self.args.seed)
        self.datasets[split] = dataset

    def build_model(self, args):
        from unicore import models
        model = models.build_model(args, self)
        return model


# Copyright (c) DP Technology.
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

import numpy as np
from itertools import product
import torch
from scipy.spatial import distance_matrix
from functools import lru_cache
from unicore.data import BaseWrapperDataset


class DistanceDataset(BaseWrapperDataset):

    def __init__(self, dataset, p=2):
        super().__init__(dataset)
        self.dataset = dataset
        self.p = p

    @lru_cache(maxsize=16)
    def __getitem__(self, idx):
        pos = self.dataset[idx].view(-1, 3).numpy()
        dist = distance_matrix(pos, pos, self.p).astype(np.float32)
        return torch.from_numpy(dist)

class GlobalDistanceDataset(BaseWrapperDataset):

    def __init__(self, dataset, lattice_matrix_dataset, p=2):
        super().__init__(dataset)
        self.dataset = dataset
        self.lattice_matrix_dataset = lattice_matrix_dataset
        self.p = p

    @lru_cache(maxsize=16)
    def __getitem__(self, idx):
        pos = self.dataset[idx].view(-1, 3).numpy()    # (n,3)
        pbc_pos = np.tile(pos, (125, 1, 1))                   # (343,n,3)
        lattice_matrix = self.lattice_matrix_dataset[idx].astype(np.float32)
        ks = [0, 1, -1, 2, -2]
        pbc_matrix = np.array(list(product(ks, repeat=3)))
        pbc_pos += np.dot(pbc_matrix, lattice_matrix).reshape(125,-1,3)
        pbc_pos = pbc_pos.reshape(-1,3)                      # (343*n,3)
        
        # dist = distance_matrix(pos, pos, self.p).astype(np.float32)
        dist_pbc = distance_matrix(pbc_pos, pos, self.p).astype(np.float32)       # (343*n,n)
        dist_pbc = (np.sort(dist_pbc.reshape(125,-1,pos.shape[0]),axis=0))[:4].transpose((1, 2, 0))    # (n,n,16)
        return torch.from_numpy(dist_pbc)

# class GlobalDistanceDataset(BaseWrapperDataset):

#     def __init__(self, dataset, lattice_matrix_dataset, p=2):
#         super().__init__(dataset)
#         self.dataset = dataset
#         self.lattice_matrix_dataset = lattice_matrix_dataset
#         self.p = p

#     @lru_cache(maxsize=16)
#     def __getitem__(self, idx):
#         pos = self.dataset[idx].view(-1, 3).numpy()    # (n,3)
#         pbc_pos = np.tile(pos, (27, 1, 1))                   # (27,n,3)
#         lattice_matrix = self.lattice_matrix_dataset[idx].astype(np.float32)
#         x = np.array([-1, 0, 1])
#         X, Y, Z = np.meshgrid(x, x, x)
#         pbc_matrix = np.stack((Y, X, Z), axis=-1).reshape(-1, 3)
#         pbc_pos += np.dot(pbc_matrix, lattice_matrix).reshape(27,-1,3)
#         pbc_pos = pbc_pos.reshape(-1,3)                      # (27*n,3)
        
#         # dist = distance_matrix(pos, pos, self.p).astype(np.float32)
#         dist_pbc = distance_matrix(pbc_pos, pos, 2).astype(np.float32)       # (27*n,n)
#         dist_pbc = np.min(dist_pbc.reshape(27,-1,pos.shape[0]), axis = 0)    # (n,n)
#         return torch.from_numpy(dist_pbc)

class EdgeTypeDataset(BaseWrapperDataset):
    def __init__(
        self,
        dataset: torch.utils.data.Dataset,
        num_types: int
    ):
        self.dataset = dataset
        self.num_types = num_types

    @lru_cache(maxsize=16)
    def __getitem__(self, index: int):
        node_input = self.dataset[index].clone()
        offset = node_input.view(-1, 1) * self.num_types + node_input.view(1, -1)
        return offset


from ai2_kit.core.util import ensure_dir, expand_globs, slice_from_str, SAMPLE_METHOD, list_sample
from ai2_kit.core.log import get_logger
from ai2_kit.domain.cp2k import dump_coord_n_cell

from typing import List, Union, Optional
from ase import Atoms
import ase.io
import random


logger = get_logger(__name__)


class AseTool:

    def __init__(self, atoms_arr: Optional[List[Atoms]] = None):
        self._atoms_arr: List[Atoms] = [] if atoms_arr is None else atoms_arr

    def read(self, *file_path_or_glob: str, nat_sort=False, **kwargs):
        """
        read atoms from file, support multiple files and glob pattern

        :param file_path_or_glob: path or glob pattern to locate data path
        :param nat_sort: sort files by natural order, default is False
        :param kwargs: other arguments for ase.io.read
        """

        files = expand_globs(file_path_or_glob, nature_sort=nat_sort)
        if len(files) == 0:
            raise FileNotFoundError(f'No file found for {file_path_or_glob}')
        for file in files:
            self._read(file, **kwargs)
        return self

    def set_cell(self, cell, scale_atoms=False, apply_constraint=True):
        for atoms in self._atoms_arr:
            atoms.set_cell(cell, scale_atoms=scale_atoms, apply_constraint=apply_constraint)
        return self

    def set_pbc(self, pbc):
        for atoms in self._atoms_arr:
            atoms.set_pbc(pbc)
        return self

    def set_by_ref(self, ref_file: str, **kwargs):
        kwargs.setdefault('index', 0)
        ref_atoms = ase.io.read(ref_file, **kwargs)
        assert isinstance(ref_atoms, Atoms), 'Only support single frame reference'
        for atoms in self._atoms_arr:
            try:
                atoms.set_cell(ref_atoms.get_cell())
            except ValueError:
                pass
            try:
                atoms.set_pbc(ref_atoms.get_pbc())
            except ValueError:
                pass
        return self

    def size(self):
        """
        size of loaded data
        """
        return len(self._atoms_arr)

    def slice(self, expr: str):
        """
        slice systems by python slice expression, for example
        `10:`, `:10`, `::2`, etc

        :param start: start index
        :param stop: stop index
        :param step: step
        """
        s = slice_from_str(expr)
        self._atoms_arr = self._atoms_arr[s]
        return self

    def shuffle(self, seed=None):
        """
        shuffle data

        :param seed: seed for random shuffle
        """
        if seed is not None:
            random.seed(seed)
        random.shuffle(self._atoms_arr)
        return self

    def sample(self, size: int, method: SAMPLE_METHOD='even', **kwargs):
        """
        sample data

        :param size: size of sample, if size is larger than data size, return all data
        :param method: method to sample, can be 'even', 'random', 'truncate', default is 'even'
        :param seed: seed for random sample, only used when method is 'random'

        Note that by default the seed is length of input list,
        if you want to generate different sample each time, you should set random seed manually
        """
        self._atoms_arr= list_sample(self._atoms_arr, size, method, **kwargs)
        return self

    def delete_atoms(self, id: Union[int, List[int]], start_id=0):
        """
        delete atoms by id or list of id

        :param id: id or list of id
        :param start_id: the start id of first item, for example, in LAMMPS the id of first item is 1, in ASE it is 0
        """

        ids = [id] if isinstance(id, int) else id
        for atoms in self._atoms_arr:
            for i in sorted(ids, reverse=True):
                assert i >= start_id, f'Invalid id {i}'
                del atoms[i - start_id]
        return self

    def write(self, filename: str, slice=None, chain=False, **kwargs):
        """
        write atoms to file
        :param filename: the filename to write
        :param slice: slice expression to select data
        :param chain: if True, return self, useful for chain operation
        :param kwargs: other arguments for ase.io.write
        """
        atoms_arr = self._atoms_arr
        if slice is not None:
            atoms_arr = atoms_arr[slice_from_str(slice)]

        ensure_dir(filename.format(i=0))
        self._write(filename, atoms_arr, **kwargs)
        if chain:
            return self

    def write_frames(self, filename: str, slice=None, chain=False, **kwargs):
        """
        write each frame to a separate file, useful to write to format only support single frame, POSCAR for example

        :param filename: the filename template, use {i} to represent the index, for example, 'frame_{i}.xyz'
        :param slice: slice expression to select data
        :param chain: if True, return self, useful for chain operation
        :param kwargs: other arguments for ase.io.write
        """
        atoms_arr = self._atoms_arr
        if slice is not None:
            atoms_arr = atoms_arr[slice_from_str(slice)]

        for i, atoms in enumerate(atoms_arr):
            _filename = filename.format(i=i)
            ensure_dir(_filename)
            self._write(_filename, atoms, **kwargs)
        if chain:
            return self

    @property
    def write_each_frame(self):
        logger.warning('write_each_frame has been deprecated, use write_frames instead')
        return self.write_frames

    def write_dplr_lammps_data(self, filename: str,
                               type_map: List[str], sel_type: List[int],
                               sys_charge_map: List[float], model_charge_map: List[float]):

        """
        write atoms to LAMMPS data file for DPLR
        the naming convention of params follows Deepmd-Kit's

        about dplr: https://docs.deepmodeling.com/projects/deepmd/en/master/model/dplr.html
        about fitting tensor: https://docs.deepmodeling.com/projects/deepmd/en/master/model/train-fitting-tensor.html


        :param filename: the filename of LAMMPS data file, use {i} to represent the index, for example, 'frame_{i}.lammps.data'
        :param type_map: the type map of atom type, for example, [O,H]
        :param sel_type: the selected type of atom, for example, [0] means atom type 0, aka O is selected
        :param sys_charge_map: the charge map of atom in system, for example, [6, 1]
        :param model_charge_map: the charge map of atom in model, for example, [-8]
        """

        from ai2_kit.domain.dplr import dump_dplr_lammps_data
        ensure_dir(filename.format(i=0))
        for i, atoms in enumerate(self._atoms_arr):
            with open(filename.format(i=i), 'w') as f:
                dump_dplr_lammps_data(f, atoms=atoms, type_map=type_map, sel_type=sel_type,
                                      sys_charge_map=sys_charge_map, model_charge_map=model_charge_map)

    def to_dpdata(self, labeled=False):
        """
        convert to dpdata format and use dpdata tool to handle

        :param labeled: if True, use dpdata.LabeledSystem, else use dpdata.System
        """
        from .dpdata import dpdata, DpdataTool
        System = dpdata.LabeledSystem if labeled else dpdata.System
        systems = [System(atom, fmt='ase/structure') for atom in self._atoms_arr]
        return DpdataTool(systems=systems)

    def _read(self, filename: str, **kwargs):
        kwargs.setdefault('index', ':')
        data = ase.io.read(filename, **kwargs)
        if not isinstance(data, list):
            data = [data]
        self._atoms_arr += data

    def _write(self, filename: str, atoms_list, **kwargs):
        assert len(atoms_list) > 0, 'No atoms to write'
        fmt = kwargs.get('format', None)
        if fmt == 'lammps-dump-text':
            return self._write_lammps_dump_text(filename, atoms_list, **kwargs)
        if fmt == 'cp2k-inc' or (fmt is None and filename.endswith('.inc')):
            if not isinstance(atoms_list, list):
                atoms_list = [atoms_list]
            if len(atoms_list) != 1:
                raise ValueError('cp2k-inc only support single frame')
            return self._write_cp2k_inc(filename, atoms_list[0])
        return ase.io.write(filename, atoms_list, **kwargs)

    def _write_cp2k_inc(self, filename: str, atoms: Atoms):
        with open(filename, 'w') as f:
            dump_coord_n_cell(f, atoms)

    def _write_lammps_dump_text(self, filename: str, atoms_list, **kwargs):
        if not isinstance(atoms_list, list):
            atoms_list = [atoms_list]
        type_map = kwargs.get('type_map')
        lines = []
        for i, atoms in enumerate(atoms_list):
            lines += [
                "ITEM: TIMESTEP", str(i),
                "ITEM: NUMBER OF ATOMS", str(len(atoms)),
            ]
            # It is kind of hacky to set boundary and cell in this way.
            # TODO: refactor it in the future.
            pbc = atoms.get_pbc()
            boundary =  ' '.join(['pp' if b else 'ff' for b in pbc])
            cell = atoms.get_cell()
            lines += [
                f"ITEM: BOX BOUNDS xy xz yz {boundary}",
                "%.16e %.16e %.16e" % (0, cell[0][0], 0),
                "%.16e %.16e %.16e" % (0, cell[1][1], 0),
                "%.16e %.16e %.16e" % (0, cell[2][2], 0),
                # TODO: support more properties, forces for example
                "ITEM: ATOMS id type x y z",
            ]

            for j, atom in enumerate(atoms):
                pos = atom.position
                no = atom.number if type_map is None else type_map.index(atom.symbol) + 1
                lines.append("%3d %2s %10.5f %10.5f %10.5f" % (j+1, no , pos[0], pos[1], pos[2]))

        with open(filename, 'w') as f:
            f.write('\n'.join(lines))



# Copyright (c) DP Technology.
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

import torch
import numpy as np
import torch.nn.functional as F
import pandas as pd
from unicore import metrics
from sklearn.metrics import r2_score, mean_squared_error, mean_absolute_error
from unicore.losses import UnicoreLoss, register_loss

ATTR_REGESTRY = {
    ### for MAP task
    ### log1p + standardization
    'hmof': [0.7352541603735866, 0.6419512242050728, 'log1p_standardization'],
    'CoRE_DB': [4.877768551522245, 1.3630433513018894, 'log1p_standardization'],
    'CoRE_MAP': [1.318703908155812, 1.657051374039756,'log1p_standardization'],
    'CoRE_MAP_CH4': [1.318703908155812, 1.657051374039756,'log1p_standardization'],
    'CoRE_MAP_CO2': [1.318703908155812, 1.657051374039756,'log1p_standardization'],
    'CoRE_MAP_Ar': [1.318703908155812, 1.657051374039756,'log1p_standardization'],
    'CoRE_MAP_Kr': [1.318703908155812, 1.657051374039756,'log1p_standardization'],
    'CoRE_MAP_Xe': [1.318703908155812, 1.657051374039756,'log1p_standardization'],
    'CoRE_MAP_O2': [1.318703908155812, 1.657051374039756,'log1p_standardization'],
    'CoRE_MAP_N2': [1.318703908155812, 1.657051374039756,'log1p_standardization'],

    'CoRE_MAP_LargeScale': [1.318703908155812, 1.657051374039756,'log1p_standardization'],
    
    'EXP_ADS': [1.318703908155812, 1.657051374039756, 'log1p_standardization'],
    'EXP_ADS_hmof': [0.7352541603735866, 0.6419512242050728, 'log1p_standardization'],

    ### single task for CoRE_O2N2
    ### log + standardization
    # 'HenrysconstantN2':[-12.554020498803377, 0.8169898691823003, 'log_standardization'], 
    # 'HenrysconstantO2':[-12.584561891924537, 0.7245019267012439, 'log_standardization'],
    # 'N2uptakemolkg':[-1.1498171780249011, 0.675196328280587, 'log_standardization'],
    # 'O2uptakemolkg':[-1.1475385472764816, 0.658851890288826, 'log_standardization'],
    # 'SelfdiffusionofN2cm2s':[-9.053232476305368, 0.9922367202982066, 'log_standardization'],
    # 'SelfdiffusionofO2cm2s':[-8.961047510953662, 0.8288661462404647, 'log_standardization'],
    # 'SelfdiffusionofN2cm2sinfDilute':[-8.947461648760754, 0.9206194145843283, 'log_standardization'],
    # 'SelfdiffusionofO2cm2sinfDilute':[-8.884236846328324, 0.7933940931623032, 'log_standardization'],

    ### rescale + log + standardization
    'HenrysconstantN2':[-1.0410950338331513, 0.8169898691823003, 'log_standardization'], 
    'HenrysconstantO2':[-1.071636426954307, 0.7245019267012439, 'log_standardization'],
    'N2uptakemolkg':[0.23647718309498947, 0.675196328280587, 'log_standardization'],
    'O2uptakemolkg':[0.23875581384340877, 0.658851890288826, 'log_standardization'],
    'SelfdiffusionofN2cm2s':[0.15710789567081626, 0.9922367202982066, 'log_standardization'],
    'SelfdiffusionofO2cm2s':[0.2492928610225201, 0.8288661462404647, 'log_standardization'],
    'SelfdiffusionofN2cm2sinfDilute':[0.26287872321543004, 0.9206194145843283, 'log_standardization'],
    'SelfdiffusionofO2cm2sinfDilute':[0.32610352564785905, 0.7933940931623032, 'log_standardization'],

    ### single task for hmof 
    #### standardization
    'CarbonDioxide_298_0.01': [0.10396354700540493, 0.22647223711266543, 'standardization'],
    'CarbonDioxide_298_0.05': [0.3552362138732747, 0.4999100688274694, 'standardization'],
    'CarbonDioxide_298_0.1': [0.5899521612050606, 0.6954814707840283, 'standardization'],
    'CarbonDioxide_298_0.5': [1.7881394809316975, 1.4182319573028714, 'standardization'],
    'CarbonDioxide_298_2.5': [4.893662963716865, 2.7665924837196814, 'standardization'],
    'Krypton_273_1.0': [0.8142831247196857, 0.49019273224067283, 'standardization'],
    'Krypton_273_10.0': [3.457146423970651, 1.844647949833803, 'standardization'],
    'Krypton_273_5.0': [2.3621157288824643, 1.1836266090761243, 'standardization'],
    'Methane_298_0.05': [0.06919052531491733, 0.11732131578468759, 'standardization'],
    'Methane_298_0.5': [0.5079206304786128, 0.48712780952084545, 'standardization'],
    'Methane_298_0.9': [0.810676617550032, 0.6673173034373471, 'standardization'],
    'Methane_298_2.5': [1.7398582577817177, 1.1031611422278471, 'standardization'],
    'Methane_298_4.5': [2.6132505457714963, 1.4531130345320122, 'standardization'],
    'Nitrogen_298_0.09': [0.030780563865632177, 0.02929089458934474, 'standardization'],
    'Nitrogen_298_0.9': [0.2797187385989161, 0.189400059544328, 'standardization'],
    'Xenon_273_1.0': [1.4689066891996336, 1.3300085891403677, 'standardization'],
    'Xenon_273_10.0': [3.9755578904359625, 2.4003920253199396, 'standardization'],
    'Xenon_273_5.0': [3.1119706247761525, 2.0242093202015337, 'standardization'],

    ### single task for CoRE_2019_N2-Ar
    ### log1p + standardization
    'N2_77_1334.0': [5.005959612727335, 0.9380200664108198, 'log1p_standardization'],
    'N2_77_4.0': [4.496992971461004, 1.1038735067142493, 'log1p_standardization'],
    'N2_77_316.0': [4.954693194444789, 0.9423899959164476, 'log1p_standardization'],
    'N2_77_23714.0': [5.040552599629131, 0.9313133637633051, 'log1p_standardization'],
    'N2_77_100000.0': [5.043832659155757, 0.9249850622789892, 'log1p_standardization'],
    'N2_77_5623.0': [5.030385999255922, 0.9362231694882166, 'log1p_standardization'],
    'N2_77_75.0': [4.856944591940231, 0.9566910917561435, 'log1p_standardization'],
    'N2_77_18.0': [4.713678987977164, 1.0070195140317848, 'log1p_standardization'],
    'N2_77_1.0': [4.230620623412422, 1.234397064939272, 'log1p_standardization'],
    'Ar_87_0.001': [1.264925291061282, 1.4867134690024557, 'log1p_standardization'],
    'Ar_87_0.01': [2.066426365172186, 1.666100149041618, 'log1p_standardization'],
    'Ar_87_0.1': [2.9871127303390046, 1.658046175780731, 'log1p_standardization'],
    'Ar_87_1.0': [3.816935011472259, 1.4580307128906818, 'log1p_standardization'],
    'Ar_87_10.0': [4.45813211359122, 1.1737996300082165, 'log1p_standardization'],
    'Ar_87_100.0': [4.905501842880746, 0.9790022884273142, 'log1p_standardization'],
    'Ar_87_1000.0': [5.1573076668377835, 0.92729736281114, 'log1p_standardization'],
    'Ar_87_2000.0': [5.202870468197515, 0.9249116774555552, 'log1p_standardization'],
    'Ar_87_4000.0': [5.241115877609449, 0.9270222084856665, 'log1p_standardization'],
    'Ar_87_6000.0': [5.259917500374311, 0.929448501473627, 'log1p_standardization'],
    'Ar_87_8000.0': [5.271761191168934, 0.9312142306996114, 'log1p_standardization'],
    'Ar_87_10000.0': [5.280589108694357, 0.9321831016872058, 'log1p_standardization'],
    'Ar_87_12000.0': [5.286852623781105, 0.9328796920711562, 'log1p_standardization'],
    'Ar_87_14000.0': [5.292341463307119, 0.9338449322238297, 'log1p_standardization'],
    'Ar_87_16000.0': [5.2972633994853835, 0.9346733153719622, 'log1p_standardization'],
    'Ar_87_18000.0': [5.3013194272697275, 0.935470730977557, 'log1p_standardization'],
    'Ar_87_20000.0': [5.3047892965617995, 0.9359557206685724, 'log1p_standardization'],
    'Ar_87_24000.0': [5.309577127060796, 0.9363774096872732, 'log1p_standardization'],
    'Ar_87_28000.0': [5.314111952981034, 0.9365075712907492, 'log1p_standardization'],
    'Ar_87_30000.0': [5.316011853742724, 0.9367891977656196, 'log1p_standardization'],
    'Ar_87_32000.0': [5.3179287846496885, 0.9366727374288051, 'log1p_standardization'],
    'Ar_87_36000.0': [5.320997819154503, 0.9366032897930457, 'log1p_standardization'],
    'Ar_87_40000.0': [5.323674724298496, 0.9368878628042351, 'log1p_standardization'],
    'Ar_87_50000.0': [5.328890590191506, 0.9368744610042196, 'log1p_standardization'],
    'Ar_87_60000.0': [5.333144071131759, 0.9369344127309929, 'log1p_standardization'],
    'Ar_87_70000.0': [5.336719947125746, 0.9366400471272878, 'log1p_standardization'],
    'Ar_87_80000.0': [5.33972128498756, 0.9365754297888167, 'log1p_standardization'],
    'Ar_87_90000.0': [5.340562553912887, 0.9372271176303055, 'log1p_standardization'],
    'Ar_87_100000.0': [5.340956577497255, 0.9371543533069057, 'log1p_standardization'],
    'N2_77_500.0': [5.037110113212742, 0.8900966851333135, 'log1p_standardization'],
    'N2_77_80.0': [4.912906318894842, 0.9148506244526978, 'log1p_standardization'],
    'N2_77_1000.0': [5.065140379743582, 0.8863932618423059, 'log1p_standardization'],
    'N2_77_200.0': [4.986376743515708, 0.9009131838615344, 'log1p_standardization'],
    'N2_77_5000.0': [5.089621749968469, 0.8889060941021609, 'log1p_standardization'],
    'N2_77_1500.0': [5.075729564131175, 0.8847394236449164, 'log1p_standardization'],
    'N2_77_10.0': [4.657726187443544, 1.0093572655049672, 'log1p_standardization'],
    'N2_77_2000.0': [5.073938058650238, 0.884034678108839, 'log1p_standardization'],
    'N2_77_20000.0': [5.106830900365691, 0.8988000188416835, 'log1p_standardization'],
    'N2_77_40000.0': [5.106670847473756, 0.8980085525936815, 'log1p_standardization'],
    'N2_77_60000.0': [5.100358262950352, 0.8914529601571912, 'log1p_standardization'],
    'N2_77_99900.0': [5.09161251433794, 0.885958368636817, 'log1p_standardization'],
    'N2_77_80000.0': [5.09540545828945, 0.8886990881959167, 'log1p_standardization'],

    ### single task for CoRE_structure_feature
    ### log1p + standardization
    # 'CoRE_PLD': [1.650712, 0.36442, 'standardization'],
    # 'CoRE_LCD': [1.963333, 0.382678, 'standardization'],
    # 'CoRE_volume': [0.386544, 0.215881, 'standardization'],    
    # 'CoRE_VF': [0.422403, 0.080336, 'standardization'],
    'hMOF_PLD': [1.865408, 0.497615, 'log1p_standardization'],
    'hMOF_LCD': [2.157661, 0.463281, 'log1p_standardization'],
    'hMOF_volume': [0.648911, 0.341842, 'log1p_standardization'],
    'hMOF_VF': [0.497774, 0.09921, 'log1p_standardization'],
    'CoRE_PLD': [4.63215803544683, 2.834787,'standardization'],
    'CoRE_LCD': [6.751438881677484, 3.905345916111568, 'standardization'],
    'CoRE_VF':[0.5305805441837245, 0.12409350367710244, 'standardization'],
    'CoRE_volume':[0.5205515104842735,0.8092927979609419, 'standardization'],


    ### single task
    ### log1p + standardization
    'Henry_C12_100K': [11.941814829776213, 5.745713382152701, 'log1p_standardization'],
    'D_C12': [1.898171e-08, 1.904578e-08, 'log1p_standardization'],
}


class Normalization(object):
    def __init__(self, mean=None, std=None, normal_type=None):
        self.mean = mean
        self.std = std
        self.normal_type = normal_type
    
    def transform(self, x):
        if self.normal_type == 'log1p_standardization':
            return (torch.log1p(x) - self.mean) / self.std
        elif self.normal_type == 'standardization':
            return (x - self.mean) / self.std
        else:
            raise ValueError('normal_type should be log1p_standardization or standardization')
    
    def inverse_transform(self, x):
        if self.normal_type == 'log1p_standardization':
            return torch.expm1(x * self.std + self.mean)
        elif self.normal_type == 'standardization':
            return x * self.std + self.mean
        else:
            raise ValueError('normal_type should be log1p_standardization or standardization')

@register_loss("mof_v1_mse")   
class MOFV1MSELoss(UnicoreLoss):
    def __init__(self, task):
        super().__init__(task)

    def forward(self, model, sample, reduce=True):
        """Compute the loss for the given sample.

        Returns a tuple with three elements:
        1) the loss
        2) the sample size, which is used as the denominator for the gradient
        3) logging outputs to display while training
        """
        net_output = model(**sample["net_input"], 
                           features_only=True, 
                           classification_head_name=self.args.classification_head_name)
        sample_size = sample["target"]["finetune_target"].size(0)
        _mean, _std, _normal_type = ATTR_REGESTRY[self.args.task_name]
        normalizer = Normalization(_mean, _std, _normal_type)
        loss = self.compute_loss(model, net_output[0], sample, normalizer, reduce=reduce)
        if not self.training:
            logging_output = {
                "loss": loss.data,
                "predict": normalizer.inverse_transform(net_output[0].view(-1, self.args.num_classes).data),
                "target": sample["target"]["finetune_target"].view(-1, self.args.num_classes).data,
                "sample_size": sample_size,
                "num_task": self.args.num_classes,
            }
        else:
            logging_output = {
                "loss": loss.data,
                "sample_size": sample_size,
                "num_task": self.args.num_classes,
            }   
        logging_output['bsz'] = sample_size          
        return loss, sample_size, logging_output

    def compute_loss(self, model, net_output, sample, normalizer, reduce=True):
        predicts = net_output.view(-1, self.args.num_classes).float()
        targets = sample['target']['finetune_target'].view(-1, self.args.num_classes).float()
        normalize_targets = normalizer.transform(targets)
        loss = F.mse_loss(
            predicts,
            normalize_targets,
            reduction="sum" if reduce else "none",
        )
        return loss

    @staticmethod
    def reduce_metrics(logging_outputs, split='valid') -> None:
        """Aggregate logging outputs from data parallel training."""
        loss_sum = sum(log.get("loss", 0) for log in logging_outputs)
        sample_size = sum(log.get("sample_size", 0) for log in logging_outputs)
        num_tasks = logging_outputs[0]["num_task"]
        # we divide by log(2) to convert the loss from base e to base 2
        metrics.log_scalar(
            "loss", loss_sum / sample_size / num_tasks, sample_size, round=3
        )
        if 'valid' in split or 'test' in split:
            predicts = torch.cat([log.get("predict") for log in logging_outputs], dim=0)
            predicts = predicts.detach().cpu().numpy()
            targets = torch.cat([log.get("target") for log in logging_outputs], dim=0)
            targets = targets.detach().cpu().numpy()
            ##### 
            r2 = r2_score(targets, predicts)
            metrics.log_scalar("{}_r2".format(split), r2, sample_size, round=3)
            

    @staticmethod
    def logging_outputs_can_be_summed(is_train) -> bool:
        """
        Whether the logging outputs returned by `forward` can be summed
        across workers prior to calling `reduce_metrics`. Setting this
        to True will improves distributed training speed.
        """
        return is_train

@register_loss("mof_v2_mse")
class MOFV2MSELoss(UnicoreLoss):
    def __init__(self, task):
        super().__init__(task)

    def forward(self, model, sample, reduce=True):
        """Compute the loss for the given sample.

        Returns a tuple with three elements:
        1) the loss
        2) the sample size, which is used as the denominator for the gradient
        3) logging outputs to display while training
        """
        net_output = model(**sample["net_input"], 
                           features_only=True, 
                           classification_head_name=self.args.classification_head_name)
        _mean, _std, _normal_type = ATTR_REGESTRY[self.args.task_name]
        normalizer = Normalization(_mean, _std, _normal_type)
        loss = self.compute_loss(model, net_output[0], sample, normalizer, reduce=reduce)
        sample_size = sample["target"]["finetune_target"].size(0)
        if not self.training:
            logging_output = {
                "loss": loss.data,
                "predict": normalizer.inverse_transform(net_output[0].view(-1, self.args.num_classes).data),
                "target": sample["target"]["finetune_target"].view(-1, self.args.num_classes).data,
                "sample_size": sample_size,
                "num_task": self.args.num_classes,
                "task_name": sample["task_name"],
            }
        else:
            logging_output = {
                "loss": loss.data,
                "sample_size": sample_size,
                "num_task": self.args.num_classes,
            }   
        logging_output['bsz'] = sample_size          
        return loss, sample_size, logging_output

    def compute_loss(self, model, net_output, sample, normalizer, reduce=True):
        predicts = net_output.view(-1, self.args.num_classes).float()
        targets = sample['target']['finetune_target'].view(-1, self.args.num_classes).float()
        normalize_targets = normalizer.transform(targets)
        loss = F.mse_loss(
            predicts,
            normalize_targets,
            reduction="sum" if reduce else "none",
        )
        return loss

    @staticmethod
    def reduce_metrics(logging_outputs, split='valid') -> None:
        """Aggregate logging outputs from data parallel training."""
        loss_sum = sum(log.get("loss", 0) for log in logging_outputs)
        sample_size = sum(log.get("sample_size", 0) for log in logging_outputs)
        num_tasks = logging_outputs[0]["num_task"]
        # we divide by log(2) to convert the loss from base e to base 2
        metrics.log_scalar(
            "loss", loss_sum / sample_size / num_tasks, sample_size, round=3
        )
        if 'valid' in split or 'test' in split:
            predicts = torch.cat([log.get("predict") for log in logging_outputs], dim=0)
            predicts = predicts.detach().cpu().numpy()
            targets = torch.cat([log.get("target") for log in logging_outputs], dim=0)
            targets = targets.detach().cpu().numpy()
            task_name_list = [item for log in logging_outputs for item in log.get("task_name")]
            predicts = predicts.reshape(-1,)
            targets = targets.reshape(-1,)
            ##### 
            r2 = r2_score(targets, predicts)
            metrics.log_scalar("{}_r2".format(split), r2, sample_size, round=4)

            if 'test' in split:
                df = pd.DataFrame({'predicts': predicts, 'targets': targets, 'task_name': task_name_list})
                if df['task_name'].nunique() < 20 and df['task_name'].nunique() > 1:
                    stats = df.groupby('task_name').apply(lambda x: r2_score(x['targets'], x['predicts'])).to_dict()
                    for task_name, r2 in stats.items():
                        metrics.log_scalar("{}_r2_{}".format(split, task_name), r2, sample_size, round=4)
            

    @staticmethod
    def logging_outputs_can_be_summed(is_train) -> bool:
        """
        Whether the logging outputs returned by `forward` can be summed
        across workers prior to calling `reduce_metrics`. Setting this
        to True will improves distributed training speed.
        """
        return is_train



from pathlib import Path
import importlib

# automatically import any Python files in the criterions/ directory
for file in sorted(Path(__file__).parent.glob("*.py")):
    if not file.name.startswith("_"):
        importlib.import_module("uninmr.losses." + file.name[:-3])


# Copyright (c) DP Technology.
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

from functools import lru_cache
from unicore.data import BaseWrapperDataset
import numpy as np
import torch


class KeyDataset(BaseWrapperDataset):
    def __init__(self, dataset, key):
        self.dataset = dataset
        self.key = key

    def __len__(self):
        return len(self.dataset)

    @lru_cache(maxsize=16)
    def __getitem__(self, idx):
        return self.dataset[idx][self.key]

class IndexDataset(BaseWrapperDataset):
    def __init__(self, dataset):
        self.dataset = dataset

    def __len__(self):
        return len(self.dataset)

    @lru_cache(maxsize=16)
    def __getitem__(self, idx):
        return idx

class ToTorchDataset(BaseWrapperDataset):
    def __init__(self, dataset, dtype='float32'):
        super().__init__(dataset)
        self.dtype = dtype

    @lru_cache(maxsize=16)
    def __getitem__(self, idx):
        d = np.array(self.dataset[idx], dtype=self.dtype)
        return torch.from_numpy(d)

class NumericalTransformDataset(BaseWrapperDataset):
    def __init__(self, dataset, ops='log1p'):
        super().__init__(dataset)
        self.dataset = dataset
        self.ops = ops

    @lru_cache(maxsize=16)
    def __getitem__(self, idx):
        if self.ops == 'log1p':
            d = np.array(self.dataset[idx], dtype='float64')
            d = np.log1p(d).astype('float32')
        else:
            d = np.array(self.dataset[idx])
        return d

class FlattenDataset(BaseWrapperDataset):
    def __init__(self, dataset):
        self.dataset = dataset
    
    def __getitem__(self, index: int):
        dd = self.dataset[index].copy()
        dd = np.array(dd).reshape(-1,)
        return dd


from .key_dataset import KeyDataset, IndexDataset, ToTorchDataset, NumericalTransformDataset, FlattenDataset
from .conformer_dataset import ConformerSampleDataset, TTADataset, TTAIndexDataset
from .normalize_dataset import NormalizeDataset, TargetScalerDataset
from .remove_hydrogen_dataset import RemoveHydrogenDataset
from .cropping_dataset import CroppingDataset
from .distance_dataset import DistanceDataset, GlobalDistanceDataset, EdgeTypeDataset
from .mask_points_dataset import MaskPointsDataset
from .pad_dataset import RightPadDataset2D0, RightPadDataset2D, RightPadDataset3D, PrependAndAppend2DDataset, PrependAndAppend3DDataset
from .lattice_dataset import LatticeNormalizeDataset, LatticeMatrixNormalizeDataset
from .lmdb_dataset import LMDBDataset, FoldLMDBDataset, StackedLMDBDataset, SplitLMDBDataset
from .select_token_dataset import SelectTokenDataset, FilterDataset
from .resample_dataset import EpochResampleDataset, EpochLogResampleDataset

__all__ = []


from typing import Optional
from ase.data import atomic_masses, atomic_numbers
import pandas as pd
import json

class LammpsTool:
    """
    A tool for LAMMPS related operations.
    """

    def analysis_density_convergence(self, in_file: str, out_file: Optional[str] = None,
                                     debug=False, err=0.0005, col_name: str = 'density'):
        """
        Select a frame by density from a property file.

        :param in_file: input properties file, e.g. properties.out
        :param out_file: output file to save the selected frame, if None, return the result
        :param col_name: column name to select by, default is 'density'
        :param debug: if True, return the average density for each frame
        :param err: error threshold for convergence, default is 0.0005

        :return: a dictionary with converged density and frames, or save to out_file
        """
        with open(in_file, 'r') as f:
            f.seek(1)  # skip the leading '#'
            df = pd.read_csv(f, delim_whitespace=True, header=0)
        if col_name not in df.columns:
            raise ValueError(f'Column {col_name} not found in {in_file}')
        density = df[col_name]

        avg_density = density.cumsum() / (df.index + 1)
        converged_density = avg_density.iloc[-1]
        converged_df = df[(density - converged_density).abs() < err]

        result = {
            'converged_density': converged_density,
            'converged_frames': converged_df.index.tolist(),
        }
        if debug:
            result['time'] = df['time'].tolist()
            result['avg_density'] = avg_density.tolist()

        if out_file:
            with open(out_file, 'w') as f:
                json.dump(result, f)
        else:
            return result

    def gen_mass_map(self, atom_types: tuple, out_file = None):
        """
        Generate a mass map for LAMMPS.
        :param atom_types: tuple of atom types, e.g. ('C', 'H', 'O')
        :return: LAMMPS mass map string
        :param out_file: output file to save the mass map, if None, return the string

        :return: mass map string, e.g.:
        ```
        mass 1 12.011 # C
        mass 2 1.008 # H
        mass 3 15.999 # O
        ```
        """
        mass_list = []
        # lookup mass from ase
        for i, atom_type in enumerate(atom_types, start=1):
            atom_no = atomic_numbers[atom_type]
            if atom_no is None:
                raise ValueError(f'Unknown atom type: {atom_type}')
            mass = atomic_masses[atom_no]
            mass_list.append(f'mass {i:>4} {mass:>8}  # {atom_type}')
        mass_map = '\n'.join(mass_list)
        if out_file:
            with open(out_file, 'w') as f:
                f.write(mass_map)
        return mass_map



# Copyright (c) DP Technology.
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

from functools import lru_cache
from unicore.data import BaseWrapperDataset
import numpy as np

class LatticeNormalizeDataset(BaseWrapperDataset):
    def __init__(self, dataset, abc, angles):
        super().__init__(dataset)
        self.dataset = dataset
        self.abc = abc
        self.angles = angles
    
    @lru_cache(maxsize=16)
    def __getitem__(self, idx):
        abc = np.array(self.dataset[idx][self.abc])
        angles = np.array(self.dataset[idx][self.angles])
        lattices = normalize(abc, angles)
        return lattices

class LatticeMatrixNormalizeDataset(BaseWrapperDataset):
    def __init__(self, dataset, lattice_matrix):
        super().__init__(dataset)
        self.dataset = dataset
        self.lattice_matrix = lattice_matrix
    
    @lru_cache(maxsize=16)
    def __getitem__(self, idx):
        lattices = np.array(self.dataset[idx][self.lattice_matrix])
        return lattices

def normalize(abc, angles):
    indices = np.argsort(abc)
    abc = abc[indices]
    angles = angles[indices]
    angles = [min(item, 180.0-item) for item in angles]
    angles = np.array(angles) / 180.0 * np.pi
    lattices = np.concatenate([abc, angles]).astype(np.float32)
    return lattices

def normalize_v2(abc, angles):
    # indices = np.argsort(abc)
    # abc = abc[indices]
    # angles = angles[indices]
    # angles = [min(item, 180.0-item) for item in angles]
    angles = np.array(angles) / 180.0 * np.pi
    # lattices = np.concatenate([abc, angles]).astype(np.float32)
    lattices = angles.astype(np.float32)
    return lattices






from argparse import Namespace
from io import StringIO
from itertools import product
from scipy.spatial import distance_matrix
import os

import numpy as np

import ase.io
from ase import Atoms

import torch
from torch.utils.data import Dataset, DataLoader

from unicore import checkpoint_utils
from unicore.data import (
    Dictionary,
    NestedDictionaryDataset,
    AppendTokenDataset,
    PrependTokenDataset,
    RightPadDataset,
    TokenizeDataset,
    RightPadDataset2D,
)

from .utils import parse_select_atom, TargetScaler
from .models import UniMatModel
from .data import (
    KeyDataset,
    IndexDataset,
    ToTorchDataset,
    DistanceDataset,
    GlobalDistanceDataset,
    EdgeTypeDataset,
    RightPadDataset3D,
    PrependAndAppend2DDataset,
    PrependAndAppend3DDataset,
    RightPadDataset2D0,
    LatticeMatrixNormalizeDataset,
    CroppingDataset,
    NormalizeDataset,
    TargetScalerDataset,
    SelectTokenDataset,
    FilterDataset,
)


from ai2_kit.core.log import get_logger
from ai2_kit.core.util import resolve_path

logger = get_logger(__name__)


class ListDataset(Dataset):
    def __init__(self, data_list):
        self.data_list = data_list

    def __len__(self):
        return len(self.data_list)

    def __getitem__(self, idx):
        return self.data_list[idx]


def extend_cells(atoms: Atoms, rcut=6):
    """
    Extend cells for periodic boundary conditions
    """

    ks = [0, 1, -1, 2, -2]
    pbc_repeat = len(ks) ** 3

    atom = atoms.get_chemical_symbols()
    pbc_atoms = atom * pbc_repeat

    pos = atoms.get_positions()
    pbc_pos = np.tile(pos, (pbc_repeat, 1, 1))
    lattice_matrix = (atoms.cell)
    pbc_matrix = np.array(list(product(ks, repeat=3)))
    pbc_pos += np.dot(pbc_matrix, lattice_matrix).reshape(pbc_repeat, -1, 3)
    pbc_pos = pbc_pos.reshape(-1, 3)
    dist_pbc = distance_matrix(pbc_pos, pos, 2).astype(np.float32)

    cells = []
    for i, e in enumerate(atom):
        dist_mask = (dist_pbc.reshape(-1, pos.shape[0])[:, i] < rcut)
        rcut_target = [0] * (len(atoms) * pbc_repeat)
        rcut_mask = [0] * (len(atoms) * pbc_repeat)
        rcut_mask[i] = 1
        cell = {
            'atoms': np.array(pbc_atoms)[dist_mask].tolist(),
            'coordinates': pbc_pos[dist_mask],
            'atoms_target': np.array(rcut_target)[dist_mask],
            'atoms_target_mask': np.array(rcut_mask)[dist_mask],
        }
        cells.append(cell)
    return cells


def get_args(model_path, dict_path, saved_dir, selected_atom='H', nmr_type='solid'):
    args = Namespace()

    args.model_path=model_path
    args.dict_path=dict_path
    args.selected_atom = selected_atom
    args.nmr_type=nmr_type
    args.saved_dir=saved_dir  # this turn out to be unused, just keep it for compatibility

    args.encoder_layers = 8
    args.encoder_embed_dim = 512
    args.encoder_ffn_embed_dim = 2048
    args.encoder_attention_heads = 64
    args.dropout = 0.1
    args.emb_dropout = 0.1
    args.attention_dropout = 0.1
    args.activation_dropout = 0.0
    args.pooler_dropout = 0.0
    args.max_seq_len = 1024
    args.activation_fn = "gelu"
    args.pooler_activation_fn = "tanh"
    args.post_ln = False
    args.masked_token_loss = -1.0
    args.masked_coord_loss = -1.0
    args.masked_dist_loss = -1.0
    args.x_norm_loss = -1.0
    args.delta_pair_repr_norm_loss = -1.0
    args.lattice_loss = -1.0
    args.encoder_layers = 15
    args.num_classes=1
    args.atom_descriptor=0
    args.classification_head_name='nmr_head'
    args.global_distance=0
    args.gaussian_kernel = True
    args.max_atoms=512
    args.max_seq_len=1024
    args.seed=1
    args.batch_size=16
    args.required_batch_size_multiple=1
    args.num_workers=8
    args.data_buffer_size=10
    args.log_format='simple'
    args.log_interval=50
    return args


def load_dataset(atoms: Atoms, args: Namespace, dictionary:Dictionary, target_scaler:TargetScaler):
    """
    Load dataset for NMRNet prediction
    """
    selected_token = parse_select_atom(dictionary, args.selected_atom)
    nmr_type = args.nmr_type

    if nmr_type == 'solid':
        atoms_info = extend_cells(atoms, rcut=6)
    elif nmr_type == 'liquid':
        ret = {
            'atoms': atoms.get_chemical_symbols(),
            'coordinates': atoms.get_positions(),
            'atoms_target': np.array([0] * len(atoms)),
            'atoms_target_mask': np.array([1] * len(atoms)),
        }
        atoms_info = [ret]
    else:
        raise ValueError(f"Invalid nmr_type: {nmr_type}")

    dataset = ListDataset(atoms_info)
    matid_dataset = IndexDataset(dataset)
    dataset = CroppingDataset(dataset, args.seed, "atoms", "coordinates", args.max_atoms)
    dataset = NormalizeDataset(dataset, "coordinates")

    token_dataset = KeyDataset(dataset, "atoms")
    token_dataset = TokenizeDataset(token_dataset, dictionary, max_seq_len=args.max_seq_len)
    atoms_target_mask_dataset = KeyDataset(dataset, "atoms_target_mask")
    select_atom_dataset = SelectTokenDataset(token_dataset=token_dataset, token_mask_dataset=atoms_target_mask_dataset, selected_token=selected_token)
    filter_list = [0 if torch.all(select_atom_dataset[i]==0) else 1 for i in range(len(select_atom_dataset))]

    dataset = FilterDataset(dataset, filter_list)
    matid_dataset = FilterDataset(matid_dataset, filter_list)
    token_dataset = FilterDataset(token_dataset, filter_list)
    select_atom_dataset = FilterDataset(select_atom_dataset, filter_list)

    coord_dataset = KeyDataset(dataset, "coordinates")

    def PrependAndAppend(dataset, pre_token, app_token):
        dataset = PrependTokenDataset(dataset, pre_token)
        return AppendTokenDataset(dataset, app_token)

    token_dataset = PrependAndAppend(token_dataset, dictionary.bos(), dictionary.eos())
    select_atom_dataset = PrependAndAppend(select_atom_dataset, dictionary.pad(), dictionary.pad())

    coord_dataset = ToTorchDataset(coord_dataset, 'float32')

    if args.global_distance:
        lattice_matrix_dataset = LatticeMatrixNormalizeDataset(dataset, 'lattice_matrix')
        logger.info("use global distance: {}".format(args.global_distance))
        distance_dataset = GlobalDistanceDataset(coord_dataset, lattice_matrix_dataset)
        distance_dataset = PrependAndAppend3DDataset(distance_dataset, 0.0)
        distance_dataset = RightPadDataset3D(distance_dataset, pad_idx=0)
    else:
        distance_dataset = DistanceDataset(coord_dataset)
        distance_dataset = PrependAndAppend2DDataset(distance_dataset, 0.0)
        distance_dataset = RightPadDataset2D(distance_dataset, pad_idx=0)

    coord_dataset = PrependAndAppend(coord_dataset, 0.0, 0.0)
    edge_type = EdgeTypeDataset(token_dataset, len(dictionary))
    tgt_dataset = KeyDataset(dataset, "atoms_target")
    tgt_dataset = TargetScalerDataset(tgt_dataset, target_scaler, args.num_classes)
    tgt_dataset = ToTorchDataset(tgt_dataset, dtype='float32')
    tgt_dataset = PrependAndAppend(tgt_dataset, dictionary.pad(), dictionary.pad())

    return NestedDictionaryDataset(
            {
                "net_input": {
                    "select_atom": RightPadDataset(
                        select_atom_dataset,
                        pad_idx=dictionary.pad(),
                    ),
                    "src_tokens": RightPadDataset(
                        token_dataset,
                        pad_idx=dictionary.pad(),
                    ),
                    "src_coord": RightPadDataset2D0(
                        coord_dataset,
                        pad_idx=0,
                    ),
                    "src_distance": distance_dataset,
                    "src_edge_type": RightPadDataset2D(
                        edge_type,
                        pad_idx=0,
                    ),
                },
                "target": {
                    "finetune_target": RightPadDataset(
                        tgt_dataset,
                        pad_idx=0,
                    ),
                },
                "matid": matid_dataset,
            },
        )


def load_model(args, dictionary):
    """
    Load model from checkpoint
    """
    state = checkpoint_utils.load_checkpoint_to_cpu(args.model_path)
    state['model'] = {
        (key.replace('classification_heads', 'node_classification_heads')
         if key.startswith('classification_heads') else key): value
        for key, value in state['model'].items()
    }
    model = UniMatModel(args, dictionary)  # type: ignore
    model.register_node_classification_head(
        args.classification_head_name,
        num_classes=args.num_classes,
        extra_dim=args.atom_descriptor,
    )
    model.load_state_dict(state["model"], strict=False)
    return model


def smiles_to_atoms(smiles):
    from rdkit import Chem
    from rdkit.Chem import AllChem
    mol = Chem.MolFromSmiles(smiles)
    mol = Chem.AddHs(mol)
    AllChem.EmbedMolecule(mol)  # type: ignore
    AllChem.UFFOptimizeMolecule(mol) # type: ignore
    symbols = [atom.GetSymbol() for atom in mol.GetAtoms()]
    positions = mol.GetConformer().GetPositions()
    atoms = Atoms(symbols=symbols, positions=positions)
    return atoms


def predict(model: UniMatModel, dataloader: DataLoader,
            classification_head_name, num_classes, target_scaler: TargetScaler):
    """
    Predict NMR Spectrum
    """
    model.eval()
    with torch.no_grad():
        all_predicts = []
        for batch in dataloader:
            net_output = model(**{k.replace('net_input.', ''): v for k, v in batch.items() if k.startswith('net_input.')},
                            features_only=True,
                            classification_head_name=classification_head_name)
            predict = target_scaler.inverse_transform(
                net_output[0].view(-1, num_classes).data.cpu()
            ).astype('float32')
            all_predicts.append(predict)
        final_predicts = np.concatenate(all_predicts)
    return final_predicts


def predict_cli(model_path: str, dict_path: str, scaler_path: str,
                selected_atom: str, nmr_type: str, use_cuda=False, cuda_device_id=None,
                smiles: str = '', data_file: str = '', data: str = '', format=None,
                return_xyz=False):
    """
    Command line interface for NMRNet prediction.

    You can provide input data with one of `data_file`, `data` or `smiles`.

    :param model_path: path to the model checkpoint, e.g 'model.pt'
    :param dict_path: path to the dictionary file, e.g 'dict.txt'
    :param scaler_path: path to the scaler file, e.g 'target_scaler.ss'
    :param selected_atom: selected atom for prediction, e.g 'H'
    :param nmr_type: type of NMR prediction, should be 'solid' or 'liquid'
    :param use_cuda: whether to use GPU for prediction, default is False
    :param cuda_device_id: GPU device id, default is None, required when use_cuda is True
    :param data_file: path to the input data file, which should be able to parse by ASE
    :param data: input data string, default is '', you can provide data directly
    :param smiles: SMILES string for prediction, default is ''
    :param format: format of the input data file, default is None, you can find the supported format in ASE: https://wiki.fysik.dtu.dk/ase/ase/io/io.html
    """
    model_path = resolve_path(model_path)
    dict_path = resolve_path(dict_path)
    scaler_path = resolve_path(scaler_path)

    if data_file:
        data_file = resolve_path(data_file)
        atoms = ase.io.read(data_file, index=0, format=format)  # type: ignore
    elif data:
        atoms = ase.io.read(StringIO(data), index=0, format=format)  # type: ignore
    elif smiles:
        atoms = smiles_to_atoms(smiles)
    else:
        raise ValueError("data_file or smiles must be provided")

    scaler_path = os.path.abspath(scaler_path)

    scaler_dir = os.path.dirname(scaler_path)
    scaler_file = os.path.basename(scaler_path)

    args = get_args(model_path, dict_path, scaler_dir,
                    selected_atom=selected_atom, nmr_type=nmr_type)
    if use_cuda:
        torch.cuda.set_device(cuda_device_id)

    dictionary = Dictionary.load(args.dict_path)
    dictionary.add_symbol("[MASK]", is_special=True)
    target_scaler = TargetScaler(scaler_dir, scaler_file)

    assert isinstance(atoms, Atoms), "data_file must be a single ASE Atoms object"
    dataset = load_dataset(atoms, args, dictionary, target_scaler)
    dataloader = DataLoader(dataset, batch_size=1, shuffle=False)

    model = load_model(args, dictionary)
    model.half()
    if use_cuda:
        model.cuda()
    else:
        model.float()
    d = predict(model, dataloader,
                classification_head_name=args.classification_head_name,
                num_classes=args.num_classes,
                target_scaler=target_scaler)
    if return_xyz:
        f = StringIO()
        ase.io.write(f, atoms, format='extxyz')
        xyz = f.getvalue()
        return d, xyz
    else:
        return d



# Copyright (c) DP Technology.
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

import torch
import numpy as np
import torch.nn as nn
import torch.nn.functional as F
import pandas as pd
import math
from unicore import metrics
from sklearn.metrics import r2_score, mean_squared_error, mean_absolute_error
from unicore.losses import UnicoreLoss, register_loss

from ..utils import TargetScaler


@register_loss("atom_regloss_mse")
class AtomRegMSELoss(UnicoreLoss):
    def __init__(self, task):
        super().__init__(task)
        # self.target_scaler = TargetScaler()
        self.target_scaler = task.target_scaler

    def forward(self, model, sample, reduce=True):
        """Compute the loss for the given sample.

        Returns a tuple with three elements:
        1) the loss
        2) the sample size, which is used as the denominator for the gradient
        3) logging outputs to display while training
        """
        net_output = model(**sample["net_input"],
                           features_only=True,
                           classification_head_name=self.args.classification_head_name)
        select_atom = sample["net_input"]["select_atom"].view(-1, 1)
        sample_size = (select_atom==1).sum()
        src_token = sample["net_input"]["src_tokens"].view(-1, 1)
        # _mean, _std, _normal_type = ATTR_REGESTRY[self.args.task_name]
        # normalizer = Normalization(_mean, _std, _normal_type)
        loss = self.compute_loss(net_output[0], sample, reduce=reduce)
        if not self.training:
            logging_output = {
                "loss": loss.data,
                "predict": self.target_scaler.inverse_transform(net_output[0].view(-1, self.args.num_classes).data.cpu()).astype('float32'),
                "target": self.target_scaler.inverse_transform((sample["target"]["finetune_target"].view(-1, self.args.num_classes))[select_atom==1].view(-1, self.args.num_classes).data.cpu()).astype('float32'),
                "src_token": src_token,
                "select_atom": select_atom,
                "sample_size": sample_size,
                "matid": sample["matid"],
                "num_task": self.args.num_classes,
                # "encoder_rep": net_output[6],
            }
            # print("predict_output", self.target_scaler.inverse_transform(net_output[0].view(-1, self.args.num_classes).data.cpu()))
            # print("target_out", (sample["target"]["finetune_target"].view(-1, self.args.num_classes))[select_atom==1].view(-1, self.args.num_classes).data.cpu())
            # print("predict", self.target_scaler.inverse_transform(net_output[0].view(-1, self.args.num_classes).data.cpu()).astype('float32'))
            # print("target", self.target_scaler.inverse_transform((sample["target"]["finetune_target"].view(-1, self.args.num_classes))[select_atom==1].view(-1, self.args.num_classes).data.cpu()).astype('float32'))
        else:
            logging_output = {
                "loss": loss.data,
                "sample_size": sample_size,
                "num_task": self.args.num_classes,
            }
        logging_output['bsz'] = sample_size
        return loss, sample_size, logging_output

    def compute_loss(self, net_output, sample, reduce=True):
        select_atom = sample["net_input"]["select_atom"].view(-1, 1)
        predicts = net_output.view(-1, self.args.num_classes).float()
        targets = sample['target']['finetune_target'].view(-1, self.args.num_classes).float()
        # normalize_targets = torch.from_numpy(self.target_scaler.transform(targets.cpu())).to(targets.device).float()
        loss = F.mse_loss(    # l1_loss mse_loss
            predicts,
            targets[select_atom==1].view(-1, self.args.num_classes),
            # reduction="sum" if reduce else "none",
        )
        return loss

    @staticmethod
    def reduce_metrics(logging_outputs, split='valid') -> None:
        """Aggregate logging outputs from data parallel training."""
        loss_sum = sum(log.get("loss", 0) for log in logging_outputs)
        sample_size = sum(log.get("sample_size", 0) for log in logging_outputs)
        num_tasks = logging_outputs[0]["num_task"]
        # we divide by log(2) to convert the loss from base e to base 2
        metrics.log_scalar(
            "loss", loss_sum / sample_size / num_tasks, sample_size, round=3
        )
        def reg_metrics(targets, predicts):
            r2 = r2_score(targets, predicts)
            mae = mean_absolute_error(targets, predicts)
            mse = mean_squared_error(targets, predicts)
            rmse = math.sqrt(mse)
            return r2, mae, mse, rmse

        if 'valid' in split or 'test' in split:
            predicts = np.concatenate([log.get("predict") for log in logging_outputs], axis=0)
            # predicts = predicts.detach().cpu().numpy()
            targets = np.concatenate([log.get("target") for log in logging_outputs], axis=0)
            # targets = targets.detach().cpu().numpy()
            #####
            r2, mae, mse, rmse = reg_metrics(targets, predicts)
            metrics.log_scalar("{}_r2".format(split), r2, sample_size, round=4)
            metrics.log_scalar("{}_mae".format(split), mae, sample_size, round=4)
            metrics.log_scalar("{}_mse".format(split), mse, sample_size, round=4)
            metrics.log_scalar("{}_rmse".format(split), rmse, sample_size, round=4)
            #####
            src_tokens = torch.cat([log.get("src_token")[log.get("select_atom")==1] for log in logging_outputs], dim=0)
            src_tokens = src_tokens.detach().cpu().numpy()
            elemenets = set(src_tokens)
            if len(elemenets) > 1:
                for element in elemenets:
                    element_targets = targets[src_tokens==element]
                    element_predicts = predicts[src_tokens==element]
                    r2, mae, mse, rmse = reg_metrics(element_targets, element_predicts)
                    element_sample_size = len(src_tokens[src_tokens==element])
                    metrics.log_scalar("{}_{}_r2".format(split, [element]), r2, element_sample_size, round=4)
                    metrics.log_scalar("{}_{}_mae".format(split, [element]), mae, element_sample_size, round=4)
                    metrics.log_scalar("{}_{}_mse".format(split, [element]), mse, element_sample_size, round=4)
                    metrics.log_scalar("{}_{}_rmse".format(split, [element]), rmse, element_sample_size, round=4)

@register_loss("atom_regloss_mae")
class AtomRegMAELoss(AtomRegMSELoss):
    def __init__(self, task):
        super().__init__(task)
        # self.target_scaler = TargetScaler()
        self.target_scaler = task.target_scaler
    # def forward(self, model, sample, reduce=True):
    #     """Compute the loss for the given sample.

    #     Returns a tuple with three elements:
    #     1) the loss
    #     2) the sample size, which is used as the denominator for the gradient
    #     3) logging outputs to display while training
    #     """
    #     net_output = model(**sample["net_input"],
    #                        features_only=True,
    #                        classification_head_name=self.args.classification_head_name)
    #     select_atom = sample["net_input"]["select_atom"].view(-1, 1)
    #     sample_size = (select_atom==1).sum()
    #     src_token = sample["net_input"]["src_tokens"].view(-1, 1)
    #     # _mean, _std, _normal_type = ATTR_REGESTRY[self.args.task_name]
    #     # normalizer = Normalization(_mean, _std, _normal_type)
    #     loss = self.compute_loss(net_output[0], sample, reduce=reduce)

    #     if not self.training:
    #         logging_output = {
    #             "loss": loss.data,
    #             "predict": (torch.from_numpy(self.target_scaler.inverse_transform(net_output[0].view(-1, self.args.num_classes).data.cpu())).to(net_output[0].device).float()),
    #             "target": (sample["target"]["finetune_target"].view(-1, self.args.num_classes).data),
    #             "src_token": src_token,
    #             "select_atom": select_atom,
    #             "sample_size": sample_size,
    #             "num_task": self.args.num_classes,
    #         }
    #         # print(logging_output["target"][:100])
    #     else:
    #         logging_output = {
    #             "loss": loss.data,
    #             "sample_size": sample_size,
    #             "num_task": self.args.num_classes,
    #         }
    #     logging_output['bsz'] = sample_size
    #     return loss, sample_size, logging_output

    def compute_loss(self, net_output, sample, reduce=True):
        select_atom = sample["net_input"]["select_atom"].view(-1, 1)
        predicts = net_output.view(-1, self.args.num_classes).float()
        targets = sample['target']['finetune_target'].view(-1, self.args.num_classes).float()
        # print("loss.py", predicts, targets)
        # normalize_targets = torch.from_numpy(self.target_scaler.transform(targets.cpu())).to(targets.device).float()
        loss = F.l1_loss(    # l1_loss mse_loss
            predicts,
            targets[select_atom==1].view(-1, self.args.num_classes),
            # reduction="sum" if reduce else "none",
        )
        # l1_loss = nn.L1Loss()
        # loss = l1_loss(    # l1_loss mse_loss
        #     predicts[select_atom==1],
        #     normalize_targets[select_atom==1],
        # )
        return loss

    # @staticmethod
    # def reduce_metrics(logging_outputs, split='valid') -> None:
    #     """Aggregate logging outputs from data parallel training."""
    #     loss_sum = sum(log.get("loss", 0) for log in logging_outputs)
    #     sample_size = sum(log.get("sample_size", 0) for log in logging_outputs)
    #     num_tasks = logging_outputs[0]["num_task"]
    #     # we divide by log(2) to convert the loss from base e to base 2
    #     metrics.log_scalar(
    #         "loss", loss_sum / sample_size / num_tasks, sample_size, round=3
    #     )
    #     def reg_metrics(targets, predicts):
    #         r2 = r2_score(targets, predicts)
    #         mae = mean_absolute_error(targets, predicts)
    #         mse = mean_squared_error(targets, predicts)
    #         rmse = math.sqrt(mse)
    #         return r2, mae, mse, rmse

    #     if 'valid' in split or 'test' in split:
    #         predicts = torch.cat([log.get("predict") for log in logging_outputs], dim=0)
    #         predicts = predicts.detach().cpu().numpy()
    #         targets = torch.cat([log.get("target") for log in logging_outputs], dim=0)
    #         targets = targets.detach().cpu().numpy()
    #         #####
    #         r2, mae, mse, rmse = reg_metrics(targets, predicts)
    #         metrics.log_scalar("{}_r2".format(split), r2, sample_size, round=4)
    #         metrics.log_scalar("{}_mae".format(split), mae, sample_size, round=4)
    #         metrics.log_scalar("{}_mse".format(split), mse, sample_size, round=4)
    #         metrics.log_scalar("{}_rmse".format(split), rmse, sample_size, round=4)
    #         #####
    #         src_tokens = torch.cat([log.get("src_token") for log in logging_outputs], dim=0)
    #         src_tokens = src_tokens.detach().cpu().numpy()
    #         elemenets = set(src_tokens)
    #         if len(elemenets) > 1:
    #             for element in elemenets:
    #                 element_targets = targets[src_tokens==element]
    #                 element_predicts = predicts[src_tokens==element]
    #                 r2, mae, mse, rmse = reg_metrics(element_targets, element_predicts)
    #                 element_sample_size = len(src_tokens[src_tokens==element])
    #                 metrics.log_scalar("{}_{}_r2".format(split, [element]), r2, element_sample_size, round=4)
    #                 metrics.log_scalar("{}_{}_mae".format(split, [element]), mae, element_sample_size, round=4)
    #                 metrics.log_scalar("{}_{}_mse".format(split, [element]), mse, element_sample_size, round=4)
    #                 metrics.log_scalar("{}_{}_rmse".format(split, [element]), rmse, element_sample_size, round=4)





# Copyright (c) DP Technology.
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

import torch
import torch.nn.functional as F
from unicore import metrics
from unicore.losses import UnicoreLoss, register_loss


@register_loss("unimat_rcut")
class UniMatRCutLoss(UnicoreLoss):
    def __init__(self, task):
        super().__init__(task)
        self.padding_idx = task.dictionary.pad()
        self.seed = task.seed

    def forward(self, model, sample, reduce=True):
        input_key = "net_input"
        target_key = "target"
        masked_tokens = sample[target_key]["tokens_target"].ne(self.padding_idx)
        select_atom = sample[input_key]["select_atom"]
        sample_size = (select_atom==1).long().sum()
        logits_encoder, encoder_distance, encoder_coord, lattice, x_norm, delta_encoder_pair_rep_norm = model(**sample[input_key], encoder_masked_tokens=(select_atom==1))
        target = sample[target_key]["tokens_target"]
        if select_atom is not None:
            target = target[select_atom==1]
        masked_token_loss = F.nll_loss(
            F.log_softmax(logits_encoder, dim=-1, dtype=torch.float32),
            target,
            ignore_index=self.padding_idx,
            reduction="mean",
        )
        masked_pred = logits_encoder.argmax(dim=-1)
        masked_hit = (masked_pred == target).long().sum()
        masked_cnt = sample_size
        loss = masked_token_loss * self.args.masked_token_loss
        if not self.training:
            logging_output = {
                "sample_size": 1,
                "bsz": sample[target_key]["tokens_target"].size(0),
                "seq_len": sample[target_key]["tokens_target"].size(1) * sample[target_key]["tokens_target"].size(0),
                "masked_token_loss": masked_token_loss.data,
                "masked_token_hit": masked_hit.data,
                "masked_token_cnt": masked_cnt,
            }
        else:
            logging_output = {
                "sample_size": 1,
                "bsz": sample[target_key]["tokens_target"].size(0),
                "seq_len": sample[target_key]["tokens_target"].size(1) * sample[target_key]["tokens_target"].size(0),
            }
        if encoder_distance is not None:
            dist_masked_tokens = (select_atom==1)
            masked_distance = encoder_distance[dist_masked_tokens, :]
            if self.args.global_distance:
                masked_distance_target = sample[target_key]["distance_target"][dist_masked_tokens][:,:,:4]
            else:
                masked_distance_target = sample[target_key]["distance_target"][dist_masked_tokens, :]
            non_pad_pos = masked_distance_target > 0
            non_pad_pos &= masked_distance_target < self.args.dist_threshold
            masked_dist_loss = F.smooth_l1_loss(
                masked_distance[non_pad_pos].view(-1).float(),
                masked_distance_target[non_pad_pos].view(-1),
                reduction="mean",
                beta=1.0,
            )
            loss = loss + masked_dist_loss * self.args.masked_dist_loss
            if not self.training:
                logging_output["masked_dist_loss"] = masked_dist_loss.data

        if encoder_coord is not None:
            coord_target = sample[target_key]["coord_target"]
            masked_coord_loss = F.smooth_l1_loss(
                encoder_coord[select_atom==1].view(-1, 3).float(),
                coord_target[select_atom==1].view(-1, 3),
                reduction="mean",
                beta=1.0,
            )
            loss = loss + masked_coord_loss * self.args.masked_coord_loss
            if not self.training:
                logging_output["masked_coord_loss"] = masked_coord_loss.data

        if lattice is not None:
            lattice_target = sample[target_key]["lattice_target"]
            # select_indices = select_atom.nonzero(as_tuple=True)
            lattice_loss = F.mse_loss(
                lattice.float(),
                lattice_target.unsqueeze(1).repeat(1, select_atom.shape[1], 1)[select_atom==1],
                reduction="mean",
            )
            loss = loss + lattice_loss * self.args.lattice_loss
            if not self.training:
                logging_output["lattice_loss"] = lattice_loss.data

        if x_norm is not None:
            loss = loss + self.args.x_norm_loss * x_norm
            if not self.training:
                logging_output["x_norm_loss"] = x_norm.data

        if delta_encoder_pair_rep_norm is not None:
            loss = loss + self.args.delta_pair_repr_norm_loss * delta_encoder_pair_rep_norm
            if not self.training:
                logging_output["delta_pair_repr_norm_loss"] = delta_encoder_pair_rep_norm.data

        logging_output["loss"] = loss.data
        return loss, 1, logging_output

    @staticmethod
    def reduce_metrics(logging_outputs, split="valid") -> None:
        """Aggregate logging outputs from data parallel training."""
        loss_sum = sum(log.get("loss", 0) for log in logging_outputs)
        sample_size = sum(log.get("sample_size", 0) for log in logging_outputs)
        metrics.log_scalar(
            "loss", loss_sum / sample_size, sample_size, round=3
        )
        bsz = sum(log.get("bsz", 0) for log in logging_outputs)
        seq_len = sum(log.get("seq_len", 0) for log in logging_outputs)
        metrics.log_scalar(
            "seq_len", seq_len / bsz, 1, round=3
        )
        if 'valid' in split or 'test' in split:
            masked_loss = sum(log.get("masked_token_loss", 0) for log in logging_outputs)
            metrics.log_scalar("masked_token_loss", masked_loss / sample_size, sample_size, round=3)

            masked_acc = sum(log.get("masked_token_hit", 0) for log in logging_outputs) / sum(log.get("masked_token_cnt", 0) for log in logging_outputs)
            metrics.log_scalar("masked_acc", masked_acc, sample_size, round=3)

            masked_coord_loss = sum(log.get("masked_coord_loss", 0) for log in logging_outputs)
            if masked_coord_loss > 0:
                metrics.log_scalar("masked_coord_loss", masked_coord_loss / sample_size, sample_size, round=3)

            masked_dist_loss = sum(log.get("masked_dist_loss", 0) for log in logging_outputs)
            if masked_dist_loss > 0:
                metrics.log_scalar("masked_dist_loss", masked_dist_loss / sample_size, sample_size, round=3)
            
            lattice_Loss = sum(log.get("lattice_loss", 0) for log in logging_outputs)
            if lattice_Loss > 0:
                metrics.log_scalar("lattice_loss", lattice_Loss / sample_size, sample_size, round=3)

            x_norm_loss = sum(log.get("x_norm_loss", 0) for log in logging_outputs)
            if x_norm_loss > 0:
                metrics.log_scalar("x_norm_loss", x_norm_loss / sample_size, sample_size, round=3)

            delta_pair_repr_norm_loss = sum(log.get("delta_pair_repr_norm_loss", 0) for log in logging_outputs)
            if delta_pair_repr_norm_loss > 0:
                metrics.log_scalar("delta_pair_repr_norm_loss", delta_pair_repr_norm_loss / sample_size, sample_size, round=3)

    @staticmethod
    def logging_outputs_can_be_summed(is_train) -> bool:
        """
        Whether the logging outputs returned by `forward` can be summed
        across workers prior to calling `reduce_metrics`. Setting this
        to True will improves distributed training speed.
        """
        return True



from collections import namedtuple
from typing import List, Tuple
from ase import Atoms

import ase.io
import os

import pandas as pd

from ai2_kit.core.util import expand_globs, slice_from_str
from ai2_kit.core.log import get_logger

logger = get_logger(__name__)


MdArray = List[Tuple[pd.DataFrame, str]]


class ModelDevi:
    """
    A tool to analyze the deviation of model from model deviation file of deepmd-kit
    """

    def __init__(self):
        self._items:List[dict] = []
        self._stats = {}

    def read(self, *dir_or_glob: str, traj_file: str, md_file='model_devi.out', ignore_error=False, **kwargs):
        """
        read model deviation from file, support multiple files and glob pattern

        :param dir_or_glob: path or glob pattern to locate data path
        :param traj_file: trajectory file name to read, relative to data path, e.g dump.lammpstrj
        :param md_file: model deviation file name to read, default is model_devi.out
        :param ignore_error: ignore error when reading files
        :param kwargs: other arguments for ase.io.read
        """
        kwargs['index'] = ':'
        dirs = expand_globs(dir_or_glob)
        if len(dirs) == 0:
            raise FileNotFoundError(f'No file found for {dir_or_glob}')
        for data_dir in dirs:
            try:
                traj_file_path = os.path.join(data_dir, traj_file)
                atoms: Atoms = ase.io.read(traj_file_path, **kwargs)  # type: ignore
                md_file_path = os.path.join(data_dir, md_file)
                with open(md_file_path, 'r') as f:
                    f.seek(1)  # skip the leading '#'
                    md_df = pd.read_csv(f, delim_whitespace=True, header=0)
                assert len(atoms) == len(md_df), 'The length of atoms and model deviation should be the same'
            except Exception as e:
                if ignore_error:
                    logger.exception(f'Error when reading {data_dir}: {e}')
                    continue
                raise e
            self._items.append({
                'atoms': atoms,
                'md_df': md_df,
                'dir': data_dir,
                'md_file': md_file_path,
                'traj_file': traj_file_path,
            })
        return self


    def grade(self, lo: float, hi: float, col: str = 'max_devi_f'):
        """
        Grade atoms based on the deviation of model into 3 levels: good, decent, poor

        the grade is based on the column of max_devi_f by default,
        if the value is below lo, the level is good,
        if the value is above hi, the level is poor,
        otherwise, the level is decent

        :param lo: the lower bound of decent level, below this value is good
        :param hi: the upper bound of decent level, above this value is poor
        :param col: the column of model deviation to grade, default is max_devi_f
        """
        for item in self._items:
            df = item['md_df']
            if col not in df.columns:
                raise ValueError(f"Unknown model deviation column: {col}")
            good = df[col] < lo
            decent = (df[col] >= lo) & (df[col] <= hi)
            poor = df[col] > hi
            self._stats[item['md_file']] = {
                'g': good.sum(),
                'd': decent.sum(),
                'p': poor.sum(),
                'all': len(df),
            }
            item['good'] = good
            item['decent'] = decent
            item['poor'] = poor
        return self

    def slice(self, expr: str):
        """
        Slice the atoms based on the expression

        :param expr: the expression to slice the atoms, e.g. '0:10'
        """
        _slice = slice_from_str(expr)
        for item in self._items:
            item['md_df'] = item['md_df'].iloc[_slice]
            # item['atoms'] = item['atoms'][_slice]
            if 'good' in item:
                item['good'] = item['good'].iloc[_slice]
            if 'decent' in item:
                item['decent'] = item['decent'].iloc[_slice]
            if 'poor' in item:
                item['poor'] = item['poor'].iloc[_slice]
        return self

    def dump_stats(self, out_file: str = '', fmt='tsv'):
        """
        Dump the statistics of grading

        :param out_file: the file path to write the statistics
        :param fmt: the format of table, default is tsv
        """
        from tabulate import tabulate

        headers = ['file', 'total', 'good', 'decent', 'poor', 'good%', 'decent%', 'poor%']
        table = []
        for file, stats in self._stats.items():
            total = stats['all']
            g = stats['g']
            d = stats['d']
            p = stats['p']
            g_pct = '{:.2%}'.format(g / total)
            d_pct = '{:.2%}'.format(d / total)
            p_pct = '{:.2%}'.format(p / total)
            table.append([file, total, g, d, p, g_pct, d_pct, p_pct])
        stats_report = tabulate(table, headers=headers, tablefmt=fmt)
        if out_file:
            with open(out_file, 'w') as f:
                f.write(stats_report)
        else:
            logger.info(f'model deviation statistics:\n{stats_report}')
        return self

    def write(self, file_path: str, inplace=False, level='decent',
              ignore_error=False, **kwargs):
        """
        Write atoms to file based on the level of grading

        :param file_path: the file path to write
        :param inplace: if True, write to the original data directory
        :param level: the level of grading to write, default is decent
        :param ignore_error: ignore error when writing files
        :param kwargs: other arguments for ase.io.write
        """
        atoms_arr = []
        for item in self._items:
            data_dir = item['dir']
            out_file = os.path.join(data_dir, file_path)
            atoms = item['atoms']
            sel = item[level]
            atoms = [atoms[i] for i in sel.index[sel]]
            if inplace:
                try: # write in place
                    ase.io.write(out_file, atoms, **kwargs)
                except Exception as e:
                    if ignore_error:
                        logger.exception(f'Error when writing {out_file}: {e}')
                    else:
                        raise e
            else:
                atoms_arr += atoms
        if not inplace:
            ase.io.write(file_path, atoms_arr, **kwargs)
        return self

    def done(self):
        return None



from rdkit import Chem
from rdkit.Chem import AllChem
import pickle
import numpy as np
import os
from tqdm import tqdm
import lmdb

from write_lmdb import write_lmdb

def get_atomic_numbers(mol):
    atomic_numbers = []  
    for atom in mol.GetAtoms():
        atomic_number = atom.GetAtomicNum()
        atomic_numbers.append(atomic_number)
    return atomic_numbers

def get_element_symbols(mol):
    element_symbols = []  
    for atom in mol.GetAtoms():
        element_symbol = atom.GetSymbol() 
        element_symbols.append(element_symbol)
    return element_symbols


mode = 'test'
ele = 'C'
file_path = f"/vepfs/fs_users/xufanjie/data/NMR/dataset/kuhn2019/{ele}.pickle"


with open(file_path, 'rb') as file:
    data = pickle.load(file)

nmrdata = data[f'{mode}_df']
nmrdata_list = []

for mol_id in nmrdata.index:
    mol = nmrdata['rdmol'][mol_id]
    
    try:
        Chem.SanitizeMol(mol)
    except:
        continue
    
    if '.' in Chem.MolToSmiles(mol):
        continue

    atoms = get_element_symbols(mol)
    assert mol.GetNumConformers() >= 1, "The molecule must have at least one conformer."
    coordinates = np.array(mol.GetConformer().GetPositions())
    atoms_target = np.array([0.0] * len(atoms))
    atoms_target_mask = np.array([0] * len(atoms))


    for atom_index, shift in nmrdata['value'][mol_id][0].items():
        atoms_target[atom_index] = shift
        atoms_target_mask[atom_index] = 1
    
    ret = {
            'atoms': atoms,
            'coordinates': coordinates,
            'atoms_target': atoms_target,
            'atoms_target_mask': atoms_target_mask,
    }

    nmrdata_list.append(ret)

write_lmdb(f'/vepfs/fs_users/xufanjie/NMRNet/data/nmrshiftdb2_2018/{ele}/{mode}.lmdb', nmrdata_list)


import numpy as np
import json
from write_lmdb import write_lmdb

def parse_geo_file(file_path):
    molecule_ids_list = []
    atoms_list = []
    coordinates_list = []

    with open(file_path, 'r') as file:
        lines = file.readlines()
        i = 0

        while i < len(lines):
            num_atoms = int(lines[i].strip())
            i += 1
            molecule_id = int(lines[i].split('_')[1])
            i += 1  

            atoms = []
            coordinates = []

            for _ in range(num_atoms):
                atom_data = lines[i].split()
                atom_type = atom_data[0]
                atom_coords = np.array([float(coord) for coord in atom_data[1:]])
                
                atoms.append(atom_type)
                coordinates.append(atom_coords)

                i += 1

            molecule_ids_list.append(molecule_id)
            atoms_list.append(atoms)
            coordinates_list.append(np.array(coordinates))

    return molecule_ids_list, atoms_list, coordinates_list


def parse_nmr_file(file_path):
    molecule_ids_list = []
    atoms_list = []
    nmrdata_list = []

    with open(file_path, 'r') as file:
        lines = file.readlines()
        i = 0

        while i < len(lines):
            num_atoms = int(lines[i].strip())
            i += 1
            molecule_id = int(lines[i].split('_')[1])
            i += 1  

            atoms = []
            nmrdata = []

            for _ in range(num_atoms):
                atom_data = lines[i].split()
                atom_type = atom_data[0]
                atom_shift = np.array([float(shift) for shift in atom_data[1:]])
                
                atoms.append(atom_type)
                nmrdata.append(atom_shift)

                i += 1

            molecule_ids_list.append(molecule_id)
            atoms_list.append(atoms)
            nmrdata_list.append(np.array(nmrdata))

    return molecule_ids_list, atoms_list, nmrdata_list


file_path = '/vepfs/fs_users/xufanjie/data/NMR/rawdata/dataset_qm9/SI_DFT_geo.xyz'
molecule_ids_geo, atoms_geo, coordinates = parse_geo_file(file_path)

file_path = '/vepfs/fs_users/xufanjie/data/NMR/rawdata/dataset_qm9/SI_DFT_NMR.txt'
molecule_ids_nmr, atoms_nmr, nmrdata = parse_nmr_file(file_path)

testid_path = "/vepfs/fs_users/xufanjie/NMRNet/data/QM9-NMR/testid.json"

with open(testid_path, 'r') as file:
    testid = json.load(file)

train_data = []
test_data = []
pahse = 0
pahsename = ['gas', 'ccl4', 'thf', 'acetone', 'methanol', 'dmso']


for i in range(len(molecule_ids_nmr)):

    ret = {
    'atoms': atoms_geo[i],
    'coordinates': np.array(coordinates[i]),
    'atoms_target': np.array(nmrdata[i][:,pahse]),
    'atoms_target_mask': np.array([1]*len(atoms_geo[i])),
    }
    if i in testid:
        test_data.append(ret)
    else:
        train_data.append(ret)

ele = 'All'
mode = 'train'
write_lmdb(f'/vepfs/fs_users/xufanjie/NMRNet/data/QM9-NMR/{pahsename[pahse]}/{ele}/{mode}.lmdb', train_data)
mode = 'valid'
write_lmdb(f'/vepfs/fs_users/xufanjie/NMRNet/data/QM9-NMR/{pahsename[pahse]}/{ele}/{mode}.lmdb', test_data)


# Copyright (c) DP Technology.
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.


import lmdb
import os
import pickle
from functools import lru_cache
from unicore.data import data_utils
import numpy as np
import logging
# import ase

logger = logging.getLogger(__name__)

class LMDBDataset:
    def __init__(self, db_path):
        self.db_path = db_path
        assert os.path.isfile(self.db_path), "{} not found".format(
            self.db_path
        )
        env = self.connect_db(self.db_path)
        with env.begin() as txn:
            self.split = os.path.splitext(os.path.basename(self.db_path))[0]
            self.dbid_file = os.path.join(os.path.dirname(self.db_path), f'{self.split}_dbid.pkl')
            if os.path.isfile(self.dbid_file):
                with open(self.dbid_file, 'rb') as f:
                    self._keys = pickle.load(f)
            else:
                self._keys = list(txn.cursor().iternext(values=False))
                with open(self.dbid_file, 'wb') as f:
                    pickle.dump(self._keys, f)

    def connect_db(self, lmdb_path, save_to_self=False):
        env = lmdb.open(
            lmdb_path,
            subdir=False,
            readonly=True,
            lock=False,
            readahead=False,
            meminit=False,
            max_readers=256,
        )
        if not save_to_self:
            return env
        else:
            self.env = env

    def __len__(self):
        return len(self._keys)

    @lru_cache(maxsize=16)
    def __getitem__(self, idx):
        if not hasattr(self, 'env'):
            self.connect_db(self.db_path, save_to_self=True)
        datapoint_pickled = self.env.begin().get(self._keys[idx])
        data = pickle.loads(datapoint_pickled)
        return data

class FoldLMDBDataset:
    def __init__(self, dataset, seed, cur_fold, nfolds=5, cache_fold_info=None):
        super().__init__()
        self.dataset = dataset
        if cache_fold_info is None:
            self.keys = []
            self.fold_start = []
            self.fold_end = []
            self.init_random_split(dataset, seed, nfolds)
        else:
            # use cache fold info
            self.keys, self.fold_start, self.fold_end = cache_fold_info
        self.cur_fold = cur_fold
        self._len = self.fold_end[cur_fold] - self.fold_start[cur_fold]
        assert len(self.fold_end) == len(self.fold_start) == nfolds

    def init_random_split(self, dataset, seed, nfolds):
        with data_utils.numpy_seed(seed):
            self.keys = np.random.permutation(len(dataset))
        average_size = (len(dataset) + nfolds - 1) // nfolds
        cur_size = 0
        for i in range(nfolds):
            self.fold_start.append(cur_size)
            cur_size = min(cur_size + average_size, len(dataset))
            self.fold_end.append(cur_size)
    
    def get_fold_info(self):
        return self.keys, self.fold_start, self.fold_end

    def __len__(self):
        return self._len

    @lru_cache(maxsize=16)
    def __getitem__(self, idx):
        global_idx = idx + self.fold_start[self.cur_fold]
        return self.dataset[self.keys[global_idx]]
            

class StackedLMDBDataset:
    def __init__(self, datasets):
        self._len = 0
        self.datasets = []
        self.idx_to_file = {}
        self.idx_offset = []
        for dataset in datasets:
            self.datasets.append(dataset)
            for i in range(len(dataset)):
                self.idx_to_file[i + self._len] = len(self.datasets) - 1
            self.idx_offset.append(self._len)
            self._len += len(dataset)

    def __len__(self):
        return self._len

    @lru_cache(maxsize=16)
    def __getitem__(self, idx):
        file_idx = self.idx_to_file[idx]
        sub_idx = idx - self.idx_offset[file_idx]
        return self.datasets[file_idx][sub_idx]
    

class SplitLMDBDataset:
    # train:valid = 9:1
    def __init__(self, dataset, seed, cur_fold, cache_fold_info= None,frac_train=0.9, frac_valid=0.1):
        super().__init__()
        self.dataset = dataset
        np.testing.assert_almost_equal(frac_train + frac_valid, 1.0)
        frac = [frac_train,frac_valid]
        if cache_fold_info is None:
            self.keys = []
            self.fold_start = []
            self.fold_end = []
            self.init_random_split(dataset, seed, frac)
        else:
            # use cache fold info
            self.keys, self.fold_start, self.fold_end = cache_fold_info
        self.cur_fold = cur_fold
        self._len = self.fold_end[cur_fold] - self.fold_start[cur_fold]
        assert len(self.fold_end) == len(self.fold_start) == 3

    def init_random_split(self, dataset, seed, frac):
        with data_utils.numpy_seed(seed):
            self.keys = np.random.permutation(len(dataset))
        frac_train,frac_valid = frac
        #average_size = (len(dataset) + nfolds - 1) // nfolds
        fold_size = [int(frac_train * len(dataset)), len(dataset)- int(frac_train * len(dataset))]
        assert sum(fold_size) == len(dataset)
        cur_size = 0
        for i in range(len(fold_size)):
            self.fold_start.append(cur_size)
            cur_size = min(cur_size + fold_size[i], len(dataset))
            self.fold_end.append(cur_size)
    
    def get_fold_info(self):
        return self.keys, self.fold_start, self.fold_end

    def __len__(self):
        return self._len

    @lru_cache(maxsize=16)
    def __getitem__(self, idx):
        global_idx = idx + self.fold_start[self.cur_fold]
        return self.dataset[self.keys[global_idx]]



# Copyright (c) DP Technology.
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

import numpy as np
from functools import lru_cache
from unicore.data import BaseWrapperDataset, data_utils


class ConformerSampleDataset(BaseWrapperDataset):
    def __init__(self, dataset, seed, atoms, coordinates):
        self.dataset = dataset
        self.seed = seed
        self.atoms = atoms
        self.coordinates = coordinates
        self.set_epoch(None)

    def set_epoch(self, epoch, **unused):
        super().set_epoch(epoch)
        self.epoch = epoch

    @lru_cache(maxsize=16)
    def __cached_item__(self, index: int, epoch: int):
        ret = self.dataset[index]
        size = len(ret[self.coordinates])
        with data_utils.numpy_seed(self.seed, epoch, index):
            sample_idx = np.random.randint(size)
        ret["coordinates"] = ret[self.coordinates][sample_idx]
        return ret

    def __getitem__(self, index: int):
        return self.__cached_item__(index, self.epoch)


class TTADataset(BaseWrapperDataset):
    def __init__(self, dataset, seed, atoms, coordinates, conf_size=10):
        self.dataset = dataset
        self.seed = seed
        self.atoms = atoms
        self.coordinates = coordinates
        self.conf_size = conf_size
        self.set_epoch(None)

    def set_epoch(self, epoch, **unused):
        super().set_epoch(epoch)
        self.epoch = epoch

    def __len__(self):
        return len(self.dataset) * self.conf_size

    @lru_cache(maxsize=16)
    def __cached_item__(self, index: int, epoch: int):
        mat_idx = index // self.conf_size
        ret = self.dataset[mat_idx]
        coord_idx = index % self.conf_size
        ret["coordinates"] =  np.array(ret[self.coordinates][coord_idx])

        return ret

    def __getitem__(self, index: int):
        return self.__cached_item__(index, self.epoch)

class TTAIndexDataset(BaseWrapperDataset):
    def __init__(self, dataset, conf_size=10):
        self.dataset = dataset
        self.conf_size = conf_size

    def __len__(self):
        return len(self.dataset) * self.conf_size

    def __getitem__(self, index: int):
        mat_idx = index // self.conf_size
        return self.dataset[mat_idx]



# Copyright (c) DP Technology.
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

import torch
import numpy as np
from functools import lru_cache
from unicore.data import BaseWrapperDataset


class SelectTokenDataset(BaseWrapperDataset):
    def __init__(
        self,
        token_dataset,
        token_mask_dataset=None,
        selected_token=np.array([-1]),
        random_choice=0,
    ):
        self.dataset = token_dataset
        self.token_mask_dataset = token_mask_dataset
        self.selected_token = selected_token
        self.random_choice = random_choice

    @lru_cache(maxsize=16)
    def __getitem__(self, index: int):
        token = self.dataset[index]
        if self.selected_token == np.array([-1]):
            ret = torch.ones_like(token)
        else:
            ret = torch.zeros_like(token)
            for val in self.selected_token:
                ret[token == val] = 1

        if self.token_mask_dataset is not None:
            token_mask = torch.from_numpy(self.token_mask_dataset[index]).long()
            ret = ret & token_mask
        if self.random_choice > 0:
            mask = torch.rand(ret.shape) < self.random_choice
            ret[mask == False] = 0
            ret[0] = 1

        return ret

# class SelectedDataset(BaseWrapperDataset):
#     def __init__(
#         self,
#         select_atom_dataset,
#     ):
#         self.dataset = select_atom_dataset

#     @lru_cache(maxsize=16)
#     def __getitem__(self, index: int):
#         return int(not torch.all(self.dataset[index] == 0))

class FilterDataset(BaseWrapperDataset):
    def __init__(
        self,
        dataset,
        filter_list
    ):
        self.dataset = dataset
        self.index_list = [index for index, value in enumerate(filter_list) if value == 1]

    def __len__(self):
        return len(self.index_list)

    @lru_cache(maxsize=16)
    def __getitem__(self, index: int):
        return self.dataset[self.index_list[index]]

